from typing import Optional, List, Union
import torch
from torch.nn.parallel import DistributedDataParallel
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.distributed.fsdp._common_utils import _FSDPState, TrainingState
from torch.distributed.fsdp.flat_param import FlatParamHandle
from torch.distributed.fsdp._runtime_utils import _post_backward_hook
import functools, weakref

def pre_pipeline_forward(num_microbatches, idx, model):
    if num_microbatches > 1 and idx == 0:
        delete_ddp_backward_hook(model)
        
def post_pipeline_forward(num_microbatches, idx, model, checkpoint_list):
    if num_microbatches > 1:
        if isinstance(model, FSDP):
            model = model._fsdp_wrapped_module
        assert(len(model)==len(checkpoint_list))
        for module, checkpoint in zip(model, checkpoint_list):
            if not checkpoint:
                if idx == num_microbatches - 1:
                    delete_fsdp_post_backward_hook(module, save_acc_grad=True, release_param=True)
                else:
                    delete_fsdp_post_backward_hook(module)
            else:
                if idx == num_microbatches - 1:
                    rewrite_fsdp_forward_no_post_backward(module)
                    
def pre_pipeline_backward(num_microbatches, idx, model, checkpoint_list):
    if num_microbatches > 1:
        if isinstance(model, FSDP):
            model = model._fsdp_wrapped_module
        assert(len(model)==len(checkpoint_list))
        if idx == num_microbatches - 1:
            register_ddp_backward_hook(model)
            for module, checkpoint in zip(model, checkpoint_list):
                if not checkpoint:
                    register_fsdp_post_backward_hook(module)
                else:
                    recover_fsdp_forward_with_post_backward(module)
    
def _register_post_backward_hooks_handle(
    state: _FSDPState,
    handle: FlatParamHandle,
) -> None:
    if not torch.is_grad_enabled():
        return
    flat_param = handle.flat_param
    already_registered = hasattr(flat_param, "_post_backward_hook_state")
    if already_registered or not flat_param.requires_grad:
        return
    # Get the `AccumulateGrad` object
    acc_grad = handle.acc_grad  # type: ignore[union-attr]
    assert acc_grad is not None
    hook_handle = acc_grad.register_hook(
        functools.partial(_post_backward_hook, state, handle)
    )
    flat_param._post_backward_hook_state = (acc_grad, hook_handle)  # type: ignore[attr-defined]

def delete_fsdp_post_backward_hook(model, save_acc_grad=False, release_param=True):
    for m in model.modules():
        if isinstance(m, FSDP):
            for handle in m._handles:
                flat_param = handle.flat_param
                if flat_param.requires_grad:
                    if hasattr(flat_param, "_post_backward_hook_state"):
                        if save_acc_grad:
                            handle.acc_grad = flat_param._post_backward_hook_state[0]
                        flat_param._post_backward_hook_state[1].remove()
                        delattr(flat_param, "_post_backward_hook_state") # whether to reduce-scatter and release grad
                    flat_param._post_backward_called = False
            if not release_param and m._is_root:
                m._post_backward_callback_queued = True # whether to release params, trades off an allgather between param memory

def register_fsdp_post_backward_hook(model):
    for m in model.modules():
        if isinstance(m, FSDP):
            for handle in m._handles:
                _register_post_backward_hooks_handle(m, handle)
            if m._is_root:
                m.training_state = TrainingState.IDLE
            m._post_backward_callback_queued = False # need to wait for post backward

def delete_ddp_backward_hook(model):
    for m in model.modules():
        # For DDP module, we need to disable gradient sync for accumulation, 
        #   and set sync manually before backward of the last microbatch.
        if isinstance(m, DDP):
            m.require_backward_grad_sync = False

def register_ddp_backward_hook(model):
    for m in model.modules():
        # For DDP module, we need to disable gradient sync for accumulation, 
        #   and set sync manually before backward of the last microbatch.
        if isinstance(m, DDP):
            m.require_forward_param_sync = True
            m.reducer.prepare_for_backward([])

def forward_delete_backward_hook(original_forward, weak_self, *args, **kwargs):
    module = weak_self()
    output = original_forward(module, *args, **kwargs)
    with torch.no_grad():
        delete_fsdp_post_backward_hook(module, save_acc_grad=True, release_param=False)
    return output

def rewrite_fsdp_forward_no_post_backward(model):
    for m in model.modules():
        if isinstance(m, FSDP):
            m.original_forward = m.forward
            m.forward = functools.partial(forward_delete_backward_hook, type(m).forward, weakref.ref(m))

def recover_fsdp_forward_with_post_backward(model):
    for m in model.modules():
        if isinstance(m, FSDP):
            m.forward = m.original_forward

def listify_model(model: Union[torch.nn.Module, List[torch.nn.Module]]) -> List[torch.nn.Module]:
    if isinstance(model, list):
        return model
    return [model]

def chunk_batch(inputs, chunks):
    if inputs is None:
        return inputs

    batches = [[] for _ in range(chunks)]
    # Actual number of chunks produced
    num_chunks = -1
    for input in inputs:
        if torch.is_tensor(input):
            # Chunk only tensors.
            tensors = input.chunk(chunks)

            # Validate number of chunks equal across all inputs.
            if num_chunks != -1 and num_chunks != len(tensors):
                raise RuntimeError(f'Found different number of chunks produced for inputs: {num_chunks} and {len(tensors)}')
            num_chunks = len(tensors)

            for i, tensor in enumerate(tensors):
                batches[i].append(tensor)
        else:
            # Replicate non-tensors or tensors wrapped with 'NoChunk'.
            for i in range(chunks):
                batches[i].append(input)
            num_chunks = chunks

    # Truncate to actual number of chunks
    batches = batches[:num_chunks]

    return batches

def unwrap_model(model, module_instances=(DistributedDataParallel,)):
    return_list = True
    if not isinstance(model, list):
        model = [model]
        return_list = False
    unwrapped_model = []
    for model_module in model:
        while isinstance(model_module, module_instances):
            model_module = model_module.module
        unwrapped_model.append(model_module)
    if not return_list:
        return unwrapped_model[0]
    return unwrapped_model


import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from torch.optim import Adam
from .dataloader import DataLoaderForLlama
import argparse
from tqdm import tqdm
import numpy as np
import random
import h5py
import time
import os
import sys
from ..utils import print_peak_memory
from megatron.initialize import initialize_megatron
from megatron import get_args
from torch.utils.data.distributed import DistributedSampler
from typing import Tuple, List
from .hybrid_parallel_model_dist import get_hybrid_parallel_configs, construct_hybrid_parallel_model, overwrite_megatron_args
from ..utils import save_profiled_memory, print_param_num
from flash_attn.models.gpt import GPTLMHeadModel
from .llama_config_utils import llama_config_to_gpt2_config, config_from_checkpoint, overwrite_configs_and_args

torch.backends.cudnn.allow_tf32 = True
torch.backends.cuda.matmul.allow_tf32 = True

def set_seed():
    seed = 123
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

def loss_func(labels, outputs):
    return outputs[0]

def forward_step_func(inputs, model):
    if isinstance(inputs, (Tuple, List)):
        outputs = model(*inputs)
    else:
        outputs = model(inputs)
    return outputs, loss_func

def train(args):
    local_rank = args.local_rank
    rank = torch.distributed.get_rank()
    torch.cuda.set_device(local_rank)
    device = torch.device("cuda", local_rank)
    world_size = torch.distributed.get_world_size()

    llama_config = config_from_checkpoint('./llama-config/', args.model_size)
    config = llama_config_to_gpt2_config(llama_config)
    overwrite_configs_and_args(config, args)
    overwrite_megatron_args(config, args)

    hybrid_parallel_configs = get_hybrid_parallel_configs(args)

    if local_rank == 0:
        print("Creating Dataloader...")
    dataset = DataLoaderForLlama(args)
    data_num_replicas = world_size // hybrid_parallel_configs['pp_deg']
    train_batch_size_input = args.global_train_batch_size // data_num_replicas
    trainloader = DataLoader(dataset=dataset,
                            batch_size=train_batch_size_input,
                            sampler=DistributedSampler(dataset,shuffle=True,num_replicas=data_num_replicas,rank=rank%data_num_replicas))

    if local_rank == 0:
        print("Creating Model...")
    
    gpt_model = GPTLMHeadModel(config, device='meta' if args.initialize_on_meta else 'cpu')
    model = construct_hybrid_parallel_model(model=gpt_model, 
                                            model_config=config, 
                                            training_args=args, 
                                            hybrid_parallel_configs=hybrid_parallel_configs)
    optimizer = Adam(model.parameters(), lr=args.lr, weight_decay=args.adam_weight_decay)
    
    if rank == 0:
        print(config)
        print_param_num(model)

    # profile_rank = 0 # profile first stage memory
    profile_ranks = [0,world_size-1]
    if args.profile and rank in profile_ranks:
        print_peak_memory("After creating model", local_rank, args.profile_type)

    start_iter, end_iter = 7, 13
     # profile first stage memory
    # profile_rank = 7 # profile last stage memory
    mem_dict = {}
    if local_rank == 0:
        print("Start training...")
    for ep in range(args.epochs):
        if not args.check_loss and not args.profile:
            trainloader = tqdm(trainloader)
        for iter, batch in enumerate(trainloader):
            start_time = time.time()
            if args.profile:
                if iter == start_iter:
                    total_start_time = start_time
                elif iter == end_iter:
                    total_end_time = start_time
                    avg_time = (total_end_time-total_start_time)/(end_iter-start_iter)
                    print("Average iteration time is: %.4f s"%avg_time)
                    return
            input_ids = batch.to(device)
            batch = [[input_ids], [input_ids]]

            if args.profile and rank in profile_ranks and iter <= 5:
                torch.cuda.reset_peak_memory_stats(local_rank)
                max_mem, cur_mem = print_peak_memory("\nBefore Forward", local_rank, args.profile_type)
                if iter <= 4:
                    mem_dict['iter_%d_before_forward'%iter] = cur_mem

            if args.pipeline_type == "gpipe":
                loss = model.gpipe_forward(forward_step_func, batch)

                if args.profile and rank in profile_ranks and iter <= 5:
                    max_mem, cur_mem = print_peak_memory("After Forward", local_rank, args.profile_type)
                    if iter <= 4:
                        mem_dict['iter_%d_after_forward'%iter] = cur_mem

                model.gpipe_backward()
            elif args.pipeline_type == "pipedream_flush":
                loss = model.pipedream_flush_forward_backward(forward_step_func, batch)
            else:
                raise NotImplementedError

            if args.profile and rank in profile_ranks and iter <= 5:
                max_mem, cur_mem = print_peak_memory("After Backward", local_rank, args.profile_type)
                if iter <= 4:
                    mem_dict['iter_%d_after_backward'%iter] = cur_mem
                    mem_dict['iter_%d_after_backward_max'%iter] = max_mem

            optimizer.step()

            if args.profile and rank in profile_ranks and iter <= 5:
                max_mem, cur_mem = print_peak_memory("After optimizer_step", local_rank, args.profile_type)

            optimizer.zero_grad()

            end_time = time.time()
            if args.check_loss or args.profile:
                if len(loss):
                    loss = np.mean([l.item() for l in loss])
                    print('[Epoch %d] (Iteration %d): Loss = %.3f'% (ep,iter,loss.item()))

            if args.profile and iter == 5:
                if rank in profile_ranks:
                    mem_dict['model_states'] = mem_dict['iter_4_after_backward']
                    if args.pipeline_type == "gpipe":
                        mem_dict['model_states_and_activation'] = mem_dict['iter_4_after_forward']
                        mem_dict['activation'] = mem_dict['iter_4_after_forward'] - mem_dict['iter_4_before_forward']
                    mem_dict['model_states_and_peak_activation'] = mem_dict['iter_4_after_backward_max']
                    mem_dict['peak_activation'] = mem_dict['iter_4_after_backward_max'] - mem_dict['iter_4_after_backward']
                    time.sleep(0.2*rank)
                    print('[Profiled memory for rank %d]:'%rank)
                    for key, val in mem_dict.items():
                        print("\t%s: %.2f MB"%(key, val))
                    if args.save_profiled_memory:
                        precision = args.mixed_precision + '_'
                        memory_config_path = './configs/memory_profiling_%dgpus_dist_%shidden%d_head%d_seqlen%d.json'%(world_size, precision, args.hidden_size, args.num_attention_heads, args.seq_length)
                        save_profiled_memory(memory_config_path, args.pp_deg, args.global_tp_deg, world_size, args.num_hidden_layers, \
                                            args.global_train_batch_size, rank, mem_dict['model_states'], mem_dict['activation'], mem_dict['peak_activation'], args.global_checkpoint)
                # return

def add_arguments(parser):
    group = parser.add_argument_group(title='our arguments')

    group.add_argument(
        "--local-rank", type=int, default=-1, help="Local rank.",
    )
    group.add_argument(
        "--global_train_batch_size", type=int, default=32, help="Global training batch size"
    )
    parser.add_argument(
        "--model_size", type=str, default='llama-7b', help="Model size.", choices=['llama-7b', 'llama-13b', 'llama-30b']
    )
    parser.add_argument(
        "--overwrite_config", type=int, default=0, help="Whether to overwrite model config"
    )
    group.add_argument(
        "--initialize_on_meta", type=int, default=1, help="Whether to initialize parameters on meta device.", choices=[0, 1]
    )
    group.add_argument(
        "--hidden_size", type=int, default=768, help="Hidden size of transformer model",
    )
    group.add_argument(
        "--num_hidden_layers", type=int, default=12, help="Number of layers"
    )
    group.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=12,
        help="Number of attention heads",
    )
    group.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    group.add_argument(
        "--vocab_size", type=int, default=30522, help="Total number of vocab"
    )
    group.add_argument(
        "--dropout_prob", type=float, default=0.1, help="Dropout rate."
    )
    group.add_argument("--max_predictions_per_seq", type=int, default=20)
    group.add_argument("-e", "--epochs", type=int,
                        default=10, help="Number of epochs")
    group.add_argument(
        "--adam_weight_decay", type=float, default=0.01, help="Weight_decay of adam"
    )

    group.add_argument(
        "--check_loss", type=int, default=0, help="Whether to check model correctness."
    )
    group.add_argument(
        "--profile", type=int, default=0, help="Whether to profile model GPU memory."
    )
    group.add_argument(
        "--save_profiled_memory", type=int, default=0, help="Whether to save profiled memory."
    )
    group.add_argument(
        "--profile_type", type=str, default='allocated', help="Profile allocated memory or reserved memory.",
        choices = ['allocated', 'reserved'],
    )
    parser.add_argument(
        "--load_params", type=int, default=0, help="Whether to load saved init params."
    )
    parser.add_argument(
        "--pp_deg", type=int, default=2, help="Pipeline parallel degree.", choices=[1,2,4,8,16,32,64],
    )
    parser.add_argument(
        "--global_tp_deg", type=int, default=-1, help="Global tensor parallel degree.", choices=[-1,1,2,4,8,16,32],
    )
    parser.add_argument(
        "--chunks", type=int, default=-1, help="Pipeline chunk num.",
    )
    parser.add_argument(
        "--global_tp_consec", type=int, default=-1, help="Global tensor parallel group consecutive flag."
    )
    parser.add_argument(
        "--fsdp", type=int, default=0, help="Apply FSDP", choices=[0, 1],
    )
    parser.add_argument(
        "--apply_strategy", type=int, default=0, help="Apply searched strategy.", choices=[0, 1],
    )
    parser.add_argument(
        "--galvatron_config_path", type=str, default=None, help="Galvatron strategy config path. If not None, galvatron will run according to json config file.",
    )
    parser.add_argument(
        "--global_checkpoint", type=int, default=0, help="Global checkpoint flag."
    )
    parser.add_argument(
        "--mixed_precision", type=str, default='bf16', help="Mixed precision option.", choices=['fp32', 'fp16', 'bf16'],
    )
    parser.add_argument(
        "--pipeline_type", type=str, default="gpipe", help="Galvatron pipeline type", choices=["gpipe","pipedream_flush"],
    )
    parser.add_argument(
        "--default_dp_type", type=str, default="zero2", help="Default data parallel type", choices=["ddp","zero2","zero3"],
    )
    return parser

if __name__ == '__main__':
    initialize_megatron(extra_args_provider=add_arguments)
    args = get_args()
    set_seed()
    train(args)




def clip_configs(model_type):
    return {
        'vit-B-16': './clip_hf_configs/CLIP-ViT-B-16.json',
        'vit-L-14': './clip_hf_configs/CLIP-ViT-L-14-laion2B-s32B-b82K.json',
        'vit-H-14': './clip_hf_configs/CLIP-ViT-H-14-laion2B-s32B-b79K.json',
        'vit-g-14': './clip_hf_configs/CLIP-ViT-g-14-laion2B-s12B-b42K.json',
        'vit-bigG-14': './clip_hf_configs/CLIP-ViT-bigG-14-laion2B-39B-b160k.json',
    }[model_type]

def obtain_main_config(config):
    main_config={
        'embed_dim': config.projection_dim,
        'vision_cfg': {
            'layer_num': config.vision_config.num_hidden_layers,
            'hidden_size': config.vision_config.hidden_size,
            'head_num': config.vision_config.num_attention_heads,
            'image_size': config.vision_config.image_size,
            'patch_size': config.vision_config.patch_size,
        },
        'text_cfg': {
            'layer_num': config.text_config.num_hidden_layers,
            'hidden_size': config.text_config.hidden_size,
            'head_num': config.text_config.num_attention_heads,
            'seq_len': config.text_config.max_position_embeddings,
            'vocab_size': config.text_config.vocab_size,
        }
    }
    return main_config

def CLIP_config(model_type):
    return {
        'vit-B-16': {
            'vision': {
                'hidden_size': 768,
                'seq_len': 197,
                'layer_num': 12,
            },
            'text': {
                'hidden_size': 512,
                'seq_len': 77,
                'layer_num': 12,
            }
        },
        'vit-L-14': {
            'vision': {
                'hidden_size': 1024,
                'seq_len': 257,
                'layer_num': 24,
            },
            'text': {
                'hidden_size': 768,
                'seq_len': 77,
                'layer_num': 12,
            }
        },
        'vit-H-14': {
            'vision': {
                'hidden_size': 1280,
                'seq_len': 257,
                'layer_num': 32,
            },
            'text': {
                'hidden_size': 1024,
                'seq_len': 77,
                'layer_num': 24,
            }
        },
        'vit-g-14': {
            'vision': {
                'hidden_size': 1408,
                'seq_len': 257,
                'layer_num': 40,
            },
            'text': {
                'hidden_size': 1024,
                'seq_len': 77,
                'layer_num': 24,
            }
        },
        'vit-bigG-14': {
            'vision': {
                'hidden_size': 1664,
                'seq_len': 257,
                'layer_num': 48,
            },
            'text': {
                'hidden_size': 1280,
                'seq_len': 77,
                'layer_num': 32,
            }
        },
    }[model_type]



def generate_strategies(gpu_num, type):
    i = 1
    total = []
    while i<=gpu_num:
        total.append(i)
        i *= 2
    if type == 'full':
        strategies = []
        for pp in total:
            for tp in total:
                if pp*tp<=gpu_num:
                    dp = gpu_num // (pp * tp) 
                    if tp==1 or tp == gpu_num/pp:
                        if dp == 1:
                            strategies.append([pp,tp,dp,{}])
                        else:
                            strategies.append([pp,tp,dp,{'fsdp':0}])
                            strategies.append([pp,tp,dp,{'fsdp':1}])
                    else:
                        strategies.append([pp,tp,dp,{'tp':0,'fsdp':0}])
                        strategies.append([pp,tp,dp,{'tp':0,'fsdp':1}])
                        strategies.append([pp,tp,dp,{'tp':1,'fsdp':0}])
                        strategies.append([pp,tp,dp,{'tp':1,'fsdp':1}])
        return strategies
    elif type == 'dp+tp':
        strategies = []
        pp = 1
        for tp in total:
            if pp*tp<=gpu_num:
                dp = gpu_num // (pp * tp) 
                if tp==1 or tp == gpu_num/pp:
                    if dp == 1:
                        strategies.append([pp,tp,dp,{}])
                    else:
                        strategies.append([pp,tp,dp,{'fsdp':0}])
                        # strategies.append([pp,tp,dp,{'fsdp':1}])
                else:
                    strategies.append([pp,tp,dp,{'tp':0,'fsdp':0}])
                    # strategies.append([pp,tp,dp,{'tp':0,'fsdp':1}])
                    strategies.append([pp,tp,dp,{'tp':1,'fsdp':0}])
                    # strategies.append([pp,tp,dp,{'tp':1,'fsdp':1}])
        return strategies
    elif type == 'dp+pp':
        strategies = []
        tp = 1
        for pp in total:
            if pp*tp<=gpu_num:
                dp = gpu_num // (pp * tp) 
                if tp==1 or tp == gpu_num/pp:
                    if dp == 1:
                        strategies.append([pp,tp,dp,{}])
                    else:
                        strategies.append([pp,tp,dp,{'fsdp':0}])
                        # strategies.append([pp,tp,dp,{'fsdp':1}])
                else:
                    strategies.append([pp,tp,dp,{'tp':0,'fsdp':0}])
                    # strategies.append([pp,tp,dp,{'tp':0,'fsdp':1}])
                    strategies.append([pp,tp,dp,{'tp':1,'fsdp':0}])
                    # strategies.append([pp,tp,dp,{'tp':1,'fsdp':1}])
        return strategies


import sys
import numpy as np
from ..utils import strategy2config, read_json_config, write_json_config, str2array, array2str
import argparse
import re
from typing import List

def format(layernum, bsz, rank, type):
    if isinstance(layernum, List):
        return "layernum[%s]_bsz%d_rank%d_%s"%(array2str(layernum), bsz, rank, type)
    else:
        return "layernum%d_bsz%d_rank%d_%s"%(layernum, bsz, rank, type)

def match_str(s):
    if '[' in s and ']' in s:
        layernum = str2array(s[s.find('[')+1:s.find(']')])
        s = s[s.find(']')+2:]
        pattern = r'bsz(\d+)_rank(\d+)_(\w+)'
        match = re.match(pattern, s)
        bsz = int(match.group(1))
        rank = int(match.group(2))
        type = match.group(3)
    else:
        pattern = r'layernum(\d+)_bsz(\d+)_rank(\d+)_(\w+)'
        match = re.match(pattern, s)
        layernum = int(match.group(1))
        bsz = int(match.group(2))
        rank = int(match.group(3))
        type = match.group(4)
    return layernum, bsz, rank, type

def get_layernums_bsz(re, layernum_list_flag):
    layernums = []
    for key in re.keys():
        layernum, bsz, _, _ = match_str(key)
        if layernum_list_flag:
            layernums.extend(layernum)
        else:
            layernums.append(layernum)
    layernums = sorted(list(set(layernums)))
    return layernums, bsz

def get_layernum_keys(layernums, layertype, layernum_list_flag, l):
    if len(layernums) == 1:
        layernum_key = [layernums[0]] * layertype if layernum_list_flag else layernums[0]
        return layernum_key, layernum_key
    if layernum_list_flag:
        layernum_key_0, layernum_key_1 = [layernums[0]] * layertype, [layernums[0]] * layertype
        layernum_key_1[l] = layernums[1]
    else:
        layernum_key_0, layernum_key_1 = layernums[0], layernums[1]
    return layernum_key_0, layernum_key_1

def total_memcost(pp_deg, layernum, layertype, layernum_list_flag, per_layer_cost, stage_idx):
    if layernum_list_flag:
        layer_costs = []
        for l in range(layertype):
            layer_costs += [per_layer_cost[l]] * layernum
        
        total_layer_num = layertype * layernum
        avg_layer_num = int(total_layer_num // pp_deg)
        last_layer_num = total_layer_num - avg_layer_num * (pp_deg-1)
        pp_divide = [avg_layer_num] * (pp_deg-1) + [last_layer_num]
        return np.sum(layer_costs[int(np.sum(pp_divide[:stage_idx])):int(np.sum(pp_divide[:stage_idx+1]))])
    else:
        return per_layer_cost * layernum / pp_deg

def process_profiled_memory(args):
    world_size = args.gpu_num
    memory_config_path = './configs/memory_profiling_%dgpus_dist_%s_%s.json'%(world_size, args.mixed_precision, args.model_type)
    config = read_json_config(memory_config_path)

    bsz = 8
    layernum_list_flag = False
    layertype = 1
    
    for key in config['1_1_%s'%world_size].keys():
        layernum, bsz, _, _ = match_str(key)
        if isinstance(layernum, List):
            layertype = len(layernum)
            layernum_list_flag = True
        break
    if layernum_list_flag:
        param_result_list, act_result_list, param_list = [dict() for _ in range(layertype)], [dict() for _ in range(layertype)], [-1]*layertype
    else:
        param_result, act_result, param = dict(), dict(), -1

    pp_deg, tp_deg = 1, 1
    while True:
        if pp_deg * tp_deg > world_size:
            break
        print(pp_deg, tp_deg)
        strategy = '%d_%d_%d'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
        if strategy not in config:
            tp_deg *= 2
            continue
        re = config[strategy]
        layernums, bsz = get_layernums_bsz(re, layernum_list_flag)
        for l in range(layertype):
            layernum_key_0, layernum_key_1 = get_layernum_keys(layernums, layertype, layernum_list_flag, l)
            param_per_layer = (re[format(layernum_key_1, bsz, 0, 'ms')] - re[format(layernum_key_0, bsz, 0, 'ms')])/(layernums[1]-layernums[0])*pp_deg/4
            act_per_layer_per_sample = (re[format(layernum_key_1, bsz, 0, 'act')] - re[format(layernum_key_0, bsz, 0, 'act')])/(layernums[1]-layernums[0])*pp_deg/(pp_deg*tp_deg)
            act_per_layer_per_sample *= world_size / bsz
            if args.dp_type == 'zero3':
                param_per_layer *= world_size//pp_deg//tp_deg
            if layernum_list_flag:
                param_result, act_result, param = param_result_list[l], act_result_list[l], param_list[l]
            param = max(param, param_per_layer*tp_deg)
            print(param_per_layer, act_per_layer_per_sample, param)
            param_result[tp_deg] = param_per_layer
            act_result[tp_deg] = act_per_layer_per_sample
            if layernum_list_flag:
                param_result_list[l], act_result_list[l], param_list[l] = param_result, act_result, param
        tp_deg *= 2

    if layernum_list_flag:
        for l in range(layertype):
            print('[layertype %d:]'%l)
            param_result, act_result, param = param_result_list[l], act_result_list[l], param_list[l]
            print('param:', param)
            # print('param_dict:', param_result)
            print('act_dict:', act_result)
    else:
        print('param:', param)
        # print('param_dict:', param_result)
        print('act_dict:', act_result)


    if layernum_list_flag:
        act_dict_c_list, act_cpt_list = [dict() for _ in range(layertype)], [-1]*layertype
    else:
        act_dict_c, act_cpt = dict(), -1
    pp_deg, tp_deg = 1, 1
    while True:
        if pp_deg * tp_deg > world_size:
            break
        print(pp_deg, tp_deg)
        strategy = '%d_%d_%d_c'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
        if strategy not in config:
            tp_deg *= 2
            continue
        re = config[strategy]
        layernums, bsz = get_layernums_bsz(re, layernum_list_flag)
        for l in range(layertype):
            layernum_key_0, layernum_key_1 = get_layernum_keys(layernums, layertype, layernum_list_flag, l)
            act_per_layer_per_sample = (re[format(layernum_key_1, bsz, 0, 'act')] - re[format(layernum_key_0, bsz, 0, 'act')])/(layernums[1]-layernums[0])*pp_deg/(pp_deg*tp_deg)
            act_per_layer_per_sample *= world_size / bsz
            print(act_per_layer_per_sample)
            if layernum_list_flag:
                act_dict_c, act_cpt = act_dict_c_list[l], act_cpt_list[l]
            act_cpt = max(act_cpt, act_per_layer_per_sample)
            act_dict_c[tp_deg] = act_per_layer_per_sample
            if layernum_list_flag:
                act_dict_c_list[l], act_cpt_list[l] = act_dict_c, act_cpt
        tp_deg *= 2

    if layernum_list_flag:
        for l in range(layertype):
            print('[layertype %d:]'%l)
            act_dict_c, act_cpt = act_dict_c_list[l], act_cpt_list[l]
            print('act_dict_c:', act_dict_c)
            print('act_cpt:', act_cpt)
            act_result_list[l]['checkpoint'] = act_cpt
    else:
        print('act_dict_c:', act_dict_c)
        print('act_cpt:', act_cpt)
        act_result['checkpoint'] = act_cpt

    inf=1e6
    other_memory_pp_off, other_memory_pp_on_first, other_memory_pp_on_last = \
        {'model_states': inf, 'activation': inf}, {'model_states': inf, 'activation': inf}, {'model_states': inf, 'activation': inf}
    pp_deg = 1
    while True:
        if pp_deg > world_size:
            break
        tp_deg = 1
        while True:
            if pp_deg * tp_deg > world_size:
                break
            print(pp_deg, tp_deg)
            strategy = '%d_%d_%d'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
            if strategy not in config:
                tp_deg *= 2
                continue
            re = config[strategy]
            layernums, bsz = get_layernums_bsz(re, layernum_list_flag)
            layernum_key, _ = get_layernum_keys(layernums, layertype, layernum_list_flag, 0)
            
            if layernum_list_flag:
                ms_cost, act_cost = [], []
                for l in range(layertype):
                    ms_cost.append(param_result_list[l][tp_deg]*4)
                    act_cost.append(act_result_list[l][tp_deg])
            else:
                ms_cost = param_result[tp_deg]
                act_cost = act_result[tp_deg]
            
            layer_ms_costs_first = total_memcost(pp_deg, layernums[0], layertype, layernum_list_flag, ms_cost, 0)
            layer_ms_costs_last = total_memcost(pp_deg, layernums[0], layertype, layernum_list_flag, ms_cost, pp_deg-1)
            layer_act_costs_first = total_memcost(pp_deg, layernums[0], layertype, layernum_list_flag, act_cost, 0)
            layer_act_costs_last = total_memcost(pp_deg, layernums[0], layertype, layernum_list_flag, act_cost, pp_deg-1)
            
            other_ms_first = re[format(layernum_key, bsz, 0, 'ms')] - layer_ms_costs_first
            if args.other_dp_type == 'zero3':
                other_ms_first = (re[format(layernum_key, bsz, 0, 'ms')] - layer_ms_costs_first / (world_size//pp_deg//tp_deg)) * world_size//pp_deg
            other_ms_last = re[format(layernum_key, bsz, world_size-1, 'ms')] - layer_ms_costs_last
            if args.other_dp_type == 'zero3':
                other_ms_last = (re[format(layernum_key, bsz, world_size-1, 'ms')] - layer_ms_costs_last / (world_size//pp_deg//tp_deg)) * world_size//pp_deg
            
            other_act_first = re[format(layernum_key, bsz, 0, 'act')] * world_size / bsz  - layer_act_costs_first * (pp_deg*tp_deg) 
            other_act_last = re[format(layernum_key, bsz, world_size-1, 'act')] * world_size / bsz - layer_act_costs_last * (pp_deg*tp_deg) 
            print(other_ms_first, other_act_first, other_ms_last, other_act_last)
            other_act_first = other_act_first if other_act_first > 0 else 0
            other_act_last = other_act_last if other_act_last > 0 else 0
            if pp_deg == 1:
                other_memory_pp_off['model_states'] = min(other_memory_pp_off['model_states'], other_ms_first)
                other_memory_pp_off['activation'] = min(other_memory_pp_off['activation'], other_act_first)
            else:
                other_memory_pp_on_first['model_states'] = min(other_memory_pp_on_first['model_states'], other_ms_first)
                other_memory_pp_on_first['activation'] = min(other_memory_pp_on_first['activation'], other_act_first / pp_deg)
                other_memory_pp_on_last['model_states'] = min(other_memory_pp_on_last['model_states'], other_ms_last)
                other_memory_pp_on_last['activation'] = min(other_memory_pp_on_last['activation'], other_act_last / pp_deg)
            tp_deg *= 2
        pp_deg *=2

    # other_memory_pp_on_first['activation'] = other_memory_pp_on_last['activation'] = max(other_memory_pp_on_first['activation'], other_memory_pp_on_last['activation'])
    print('other_memory_pp_off:', other_memory_pp_off)
    print('other_memory_pp_on_first:', other_memory_pp_on_first)
    print('other_memory_pp_on_last:', other_memory_pp_on_last)

    if layernum_list_flag:
        for l in range(layertype):
            if 'layertype_%d'%l not in config.keys():
                config['layertype_%d'%l] = dict()
            config['layertype_%d'%l]['parameter_size'] = param_list[l]
            config['layertype_%d'%l]['tp_activation_per_bsz_dict'] = act_result_list[l]
    else:
        config['parameter_size'] = param
        config['tp_activation_per_bsz_dict'] = act_result
    config['other_memory_pp_off'] = other_memory_pp_off
    config['other_memory_pp_on_first'] = other_memory_pp_on_first
    config['other_memory_pp_on_last'] = other_memory_pp_on_last
    write_json_config(config, memory_config_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--layer_num", type=int, default=24, help="Number of layers"
    )
    parser.add_argument(
        "--gpu_num", type=int, default=8, help="Number of GPUs",
    )
    parser.add_argument(
        "--model_type", type=str, default='vit-L-14', help="CLIP model type."
    )
    parser.add_argument(
        "--mixed_precision", type=str, default='fp32', help="Profile memory in mixed precision mode.", choices=['fp32', 'fp16', 'bf16']
    )
    parser.add_argument(
        "--dp_type", type=str, default='ddp', help="Use ddp or zero3 to profile memory.", choices=['ddp', 'zero3']
    )
    parser.add_argument(
        "--other_dp_type", type=str, default='ddp', help="Use ddp or zero3 for embeddings and cls to profile memory.", choices=['ddp', 'zero3']
    )
    args = parser.parse_args()
    process_profiled_memory(args)


import torch
from .allgather_utils import gather_from_group

class CommGroup(object):
    def __init__(self, ranks):
        assert isinstance(ranks, list) or isinstance(ranks, range), 'Rank list or range should be provided to create a CommGroup!'
        self.ranks = sorted(list(set(list(ranks))))
        self.size = len(self.ranks)
        self.group = torch.distributed.new_group(self.ranks)
    def has_rank(self, rank):
        if rank in self.ranks:
            self.intra_group_id = self.ranks.index(rank)
            return True
        return False
    def allgather(self, input):
        return gather_from_group(input, self.group)
    def print(self):
        print(self.ranks, end = ' ')

class SliceFunc(object):
    def __init__(self, slice_num, local_rank):
        self.n = slice_num
        self.local_rank = local_rank
        assert(local_rank < slice_num)
    def __call__(self, input):
        length = len(input)
        step = int(length // self.n)
        return input[int(self.local_rank*step):int((self.local_rank+1)*step)]
    def print(self):
        print("%d/%d"%(self.local_rank, self.n), end = ' ')

def show_groups(groups):
    for group in groups:
        if group is None:
            print('None', end = ' ')
        else:
            group.print()
    print()

def gen_tp_group(tp_size, to_print = True, consecutive = True):
    rank = torch.distributed.get_rank()
    world_size = torch.distributed.get_world_size()
    all_tp_groups = []

    if consecutive:
        for i in range(world_size // tp_size):
            ranks = range(i * tp_size, (i+1) * tp_size)
            group = CommGroup(ranks)
            all_tp_groups.append(group)
            if group.has_rank(rank):
                tp_group = group
    else:
        start_rank = 0
        end_rank = world_size
        dp_size = world_size // tp_size
        for j in range(dp_size):
            ranks = range(start_rank + j, end_rank, dp_size)
            group = CommGroup(ranks)
            all_tp_groups.append(group)
            if group.has_rank(rank):
                tp_group = group
    
    if rank == 0 and to_print:
        print("TP groups:", end = ' ')
        show_groups(all_tp_groups)
    return tp_group

def gen_dp_group(tp_size, to_print = True, consecutive = False):
    rank = torch.distributed.get_rank()
    world_size = torch.distributed.get_world_size()
    start_rank = 0
    end_rank = world_size
    all_dp_groups = []

    if not consecutive:
        for j in range(tp_size):
            ranks = range(start_rank + j, end_rank, tp_size)
            group = CommGroup(ranks)
            all_dp_groups.append(group)
            if group.has_rank(rank):
                dp_group = group
    else:
        dp_size = world_size // tp_size
        for i in range(tp_size):
            ranks = range(i * dp_size, (i+1) * dp_size)
            group = CommGroup(ranks)
            all_dp_groups.append(group)
            if group.has_rank(rank):
                dp_group = group
    
    if rank == 0 and to_print:
        print("DP groups:", end = ' ')
        show_groups(all_dp_groups)
    return dp_group

def gen_allgather_group(tp_size_old, tp_size_new, tp_consec_old, tp_consec_new, tp_group_new):
    world_size = torch.distributed.get_world_size()
    case0 = (tp_size_new > tp_size_old)
    case1 = (tp_size_new == tp_size_old and tp_consec_new != tp_consec_old)
    case2 = (world_size == 8 and tp_size_old == 4 and tp_size_new == 2 and tp_consec_new != tp_consec_old)
    if case0 or case1 or case2:
        return tp_group_new
    return None

def gen_slice_func(tp_size_old, tp_size_new, tp_consec_old, tp_consec_new, tp_group_old):
    world_size = torch.distributed.get_world_size()
    case0 = (tp_size_new > tp_size_old)
    case1 = (tp_size_new == tp_size_old and tp_consec_new != tp_consec_old)
    case2 = (world_size == 8 and tp_size_old == 4 and tp_size_new == 2 and tp_consec_new != tp_consec_old)
    if case0 or case1 or case2:
        slice_num = tp_size_old
        local_rank = tp_group_old.intra_group_id
        return SliceFunc(slice_num, local_rank)
    elif tp_size_new < tp_size_old:
        slice_num = tp_size_old // tp_size_new
        if tp_consec_new and tp_consec_old:
            local_rank = (torch.distributed.get_rank() // tp_size_new) % slice_num
        else:
            local_rank = (torch.distributed.get_rank() % (world_size // tp_size_new)) // (world_size // tp_size_old)
        return SliceFunc(slice_num, local_rank)
    return None

def get_dp_sizes_from_tp_sizes(all_tp_sizes):
    all_dp_sizes = []
    world_size = torch.distributed.get_world_size()
    for i in range(len(all_tp_sizes)):
        all_dp_sizes.append(world_size // all_tp_sizes[i])
    return all_dp_sizes

def get_tp_group_dict(all_tp_sizes, consecutive = True):
    tp_sizes_set = list(set(all_tp_sizes))
    tp_group_dict={}
    for tp_size in tp_sizes_set:
        tp_group_dict[tp_size] = gen_tp_group(tp_size, to_print=False, consecutive=consecutive)
    return tp_group_dict

def get_dp_group_dict(all_tp_sizes, consecutive = False):
    tp_sizes_set = list(set(all_tp_sizes))
    dp_group_dict={}
    for tp_size in tp_sizes_set:
        dp_group_dict[tp_size] = gen_dp_group(tp_size, to_print=False, consecutive=consecutive)
    return dp_group_dict

def gen_groups(all_tp_sizes, tp_consecutive_flags, show_rank = -1):
    world_size = torch.distributed.get_world_size()
    for i in range(len(all_tp_sizes)):
        tp_consec = tp_consecutive_flags[i]
        assert tp_consec == 0 or tp_consec == 1
        if all_tp_sizes[i] in [1, world_size]:
            tp_consecutive_flags[i] = 1
    tp_groups = []
    dp_groups = []
    allgather_groups = [None]
    slice_funcs = [None]
    tp_group_dict_consec = get_tp_group_dict(all_tp_sizes, True)
    tp_group_dict_inconsec = get_tp_group_dict(all_tp_sizes, False)
    dp_group_dict_consec = get_dp_group_dict(all_tp_sizes, True)
    dp_group_dict_inconsec = get_dp_group_dict(all_tp_sizes, False)
    for i in range(len(all_tp_sizes)):
        if tp_consecutive_flags[i]:
            tp_groups.append(tp_group_dict_consec[all_tp_sizes[i]])
            dp_groups.append(dp_group_dict_inconsec[all_tp_sizes[i]])
        else:
            tp_groups.append(tp_group_dict_inconsec[all_tp_sizes[i]])
            dp_groups.append(dp_group_dict_consec[all_tp_sizes[i]])
    for i in range(1, len(all_tp_sizes)):
        allgather_groups.append(gen_allgather_group(all_tp_sizes[i-1], all_tp_sizes[i], tp_consecutive_flags[i-1], tp_consecutive_flags[i], tp_groups[i]))
    for i in range(1, len(all_tp_sizes)):
        slice_funcs.append(gen_slice_func(all_tp_sizes[i-1], all_tp_sizes[i], tp_consecutive_flags[i-1], tp_consecutive_flags[i], tp_groups[i-1]))
    if show_rank >= 0 and torch.distributed.get_rank() == show_rank:
        print('====================== Galvatron Communication Group ===========================')
        print("TP groups for rank %d (all layers):"%show_rank)
        show_groups(tp_groups)
        print("DP groups for rank %d (all layers):"%show_rank)
        show_groups(dp_groups)
        # print("AllGather groups for rank %d:"%show_rank)
        # show_groups(allgather_groups)
        # print("Slice Funcs for rank %d:"%show_rank)
        # show_groups(slice_funcs)
        print('================================================================================')
    return tp_groups, dp_groups, allgather_groups, slice_funcs


import torch
from torch import nn
from torch import Tensor, device
from typing import Tuple
import sys
from megatron.model.utils import init_method_normal, scaled_init_method_normal
from megatron import get_args
from megatron.model.enums import AttnMaskType, AttnType
from megatron.model import MegatronModule
from megatron.core import mpu, tensor_parallel
import torch.nn.functional as F

class LlamaParallelMLP(MegatronModule):
    def __init__(self, init_method, 
                output_layer_init_method, 
                act_func = 'silu', 
                bias = False, 
                dropout_prob = 0.0, 
                hidden_size=None, 
                intermediate_size=None,
                tp_group = None,
                ):
        super().__init__()
        args = get_args()
        self.bias = bias

        hidden_size = args.hidden_size
        intermediate_size = int(8 * hidden_size / 3)
        intermediate_size = args.multiple_of * ((intermediate_size + args.multiple_of - 1) // args.multiple_of)

        self.w1 = tensor_parallel.ColumnParallelLinear(
            hidden_size,
            intermediate_size,
            gather_output=False,
            init_method=init_method,
            bias=bias,
            tp_group=tp_group
        )

        self.w2 = tensor_parallel.RowParallelLinear(
            intermediate_size,
            hidden_size,
            input_is_parallel=True,
            init_method=output_layer_init_method,
            bias=bias,
            tp_group=tp_group
        )

        self.w3 = tensor_parallel.ColumnParallelLinear(
            hidden_size,
            intermediate_size,
            gather_output=False,
            init_method=init_method,
            bias=bias,
            tp_group=tp_group
        )
        
        assert act_func == 'silu'
        self.activation_func = F.silu

    def forward(self, hidden_states):
        return self.w2(self.activation_func(self.w1(hidden_states)[0]) * self.w3(hidden_states)[0])[0]
    
class LlamaMLP_tp(nn.Module):
    def __init__(self, config, tp_group = None):
        super().__init__()
        args=get_args()
        init_method = init_method_normal(args.init_method_std)
        scaled_init_method = scaled_init_method_normal(args.init_method_std, args.num_layers)
        self.tp_group = tp_group.group if tp_group is not None else None
        self.mlp = LlamaParallelMLP(init_method, scaled_init_method, tp_group = self.tp_group)
        # self.post_attention_layernorm = LlamaRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
    
    def forward(self, hidden_states):
        # residual = hidden_states
        # hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        # hidden_states = residual + hidden_states
        return hidden_states


import sys
from ..utils import MemoryCostModelDist, TimeCostModelDist_with_overlap
from ..utils import DpOnModel_dist, print_strategies, form_strategy
import numpy as np
from ..utils import strategy2config, read_json_config, write_json_config, read_allreduce_bandwidth_config, read_p2p_bandwidth_config, estimate_bsz_start_16gpus, array2str
from ..utils import generate_strategies, check_optimal_chunks
import argparse
import copy

def read_profiling_configs(gpu_num):
    allreduce_bandwidth_config_path = '../env_configs/allreduce_bandwidth_dist_%d_gpus.json'%gpu_num
    comm_coe_dict = read_allreduce_bandwidth_config(allreduce_bandwidth_config_path, gpu_num=gpu_num)
    overlap_coe_path = '../env_configs/overlap_coefficient.json'
    overlap_coe = read_json_config(overlap_coe_path)['overlap_coe']
    fwd_profiling_path = './configs/forward_profiling_config.json'
    fwd_time = read_json_config(fwd_profiling_path)
    env_config_path = '../env_configs/p2p_bandwidth_dist_%d_gpus.json'%gpu_num
    p2p_comm_coe_dict = None
    # p2p_comm_coe_dict = read_p2p_bandwidth_config(env_config_path, gpu_num=gpu_num)
    return comm_coe_dict, overlap_coe, fwd_time, p2p_comm_coe_dict

def parallelism_optimization(args):
    # # only for memory checking
    # strategies = [[1,1,8,{'fsdp':0}],[1,1,8,{'fsdp':1}],
    #             [1,2,4,{'tp':0,'fsdp':0}],[1,2,4,{'tp':1,'fsdp':0}],[1,2,4,{'tp':0,'fsdp':1}],[1,2,4,{'tp':1,'fsdp':1}],
    #             [1,4,2,{'tp':0,'fsdp':0}],[1,4,2,{'tp':1,'fsdp':0}],[1,4,2,{'tp':0,'fsdp':1}],[1,4,2,{'tp':1,'fsdp':1}],
    #             [1,8,1,{}],
    #             [2,1,4,{'fsdp':0}],[2,1,4,{'fsdp':1}],
    #             [2,2,2,{'tp':0,'fsdp':0}],[2,2,2,{'tp':1,'fsdp':0}],[2,2,2,{'tp':0,'fsdp':1}],[2,2,2,{'tp':1,'fsdp':1}],
    #             [2,4,1,{}],
    #             [4,1,2,{'fsdp':0}],[4,1,2,{'fsdp':1}],
    #             [4,2,1,{}],
    #             [8,1,1,{}]]

    assert(not(args.disable_sdp and args.disable_dp))
    strategies = generate_strategies(args.gpu_num, args.type)

    strategies_new = []
    for s in strategies:
        if args.disable_tp_consec and 'tp' in s[-1] and s[-1]['tp'] == 0:
            continue
        if args.disable_sdp and 'fsdp' in s[-1] and s[-1]['fsdp'] == 1:
            continue
        if args.disable_dp and not ('fsdp' in s[-1] and s[-1]['fsdp'] == 1):
            continue
        if s[1] > 8:
            continue
        if s[0] > 4:
            continue
        strategies_new.append(s)
    strategies = strategies_new

    if args.checkpoint == 1:
        strategies_cpt = []
        for s in strategies:
            s_cpt = copy.deepcopy(s)
            s_cpt[-1]['cpt']=1
            strategies_cpt.append(s_cpt)
        strategies += strategies_cpt
        # strategies = strategies_cpt


    # Load profiling configs
    hidden_size, seq_len, layer_num, head_num = args.hidden_size, args.seq_length, args.layer_num, args.num_attention_heads
    # overwrite configs from args
    layer_num_list = [layer_num]
    gpu_num = args.gpu_num
    comm_coe_dict, overlap_coe, fwd_time, p2p_comm_coe_dict = read_profiling_configs(gpu_num)
    bf16 = '_bf16' if args.mixed_precision else ''
    fwd_time = fwd_time['fwd_time%s_hidden_%d'%(bf16,hidden_size)]


    print('================================================================================')
    print('------- Model configs -------')
    print('Layer_num:', layer_num)
    print('Hidden_size:', hidden_size)
    print('================================================================================')
    print('--- Optimization configs ----')
    print('Memory_constraint: %d GB'%args.memory_constraint)
    print('Optimization type:', args.type)
    print('================================================================================')
    print('---- Environment configs ----')
    print('Allreduce comm_coe dict (ms/MB):', comm_coe_dict)
    print('P2P comm_coe dict (ms/MB):', p2p_comm_coe_dict)
    print('Overlap coefficient:', overlap_coe)
    print('--- Model forward configs ---')
    print('Forward computation time:', fwd_time)
    print('================================================================================')

    def optimal_chunk_func(local_bsz, strategy):
        if strategy[0] == 1:
            return 1
        local_bsz = local_bsz // strategy[1]
        re = np.ceil(local_bsz / 4)
        re = 1 if re == 0 else re
        re = min(8,re)
        return int(re)

    microbatch = True
    forward_compute_time_per_layer = fwd_time
    bf16 = '_bf16' if args.mixed_precision else ''
    memory_config_path = './configs/memory_profiling_%dgpus_dist%s_hidden%d_head%d_seqlen%d.json'%(8, bf16, hidden_size, head_num, seq_len)
    mem_config = read_json_config(memory_config_path)
    parameter_size = mem_config['parameter_size']
    tp_activation_per_bsz_dict = mem_config['tp_activation_per_bsz_dict'].copy()
    for key, val in mem_config['tp_activation_per_bsz_dict'].items():
        if len(key) < 5:
            tp_activation_per_bsz_dict[int(key)] = val
    other_memory_pp_off = mem_config['other_memory_pp_off']
    other_memory_pp_on_first, other_memory_pp_on_last = mem_config['other_memory_pp_on_first'], mem_config['other_memory_pp_on_last']
    other_memory_pp_on = {'first_stage':other_memory_pp_on_first, 'last_stage':other_memory_pp_on_last}
    print(parameter_size)
    print(tp_activation_per_bsz_dict)
    print(other_memory_pp_off)
    print(other_memory_pp_on)
    memcost_model_args = {  'parameter_size': parameter_size,
                            'tp_activation_per_bsz_dict': tp_activation_per_bsz_dict,
                            'other_memory_pp_off': other_memory_pp_off,
                            'other_memory_pp_on': other_memory_pp_on,
                            'microbatch': microbatch,
                            'optimal_chunk_func': optimal_chunk_func,
                            'model_type': 'chatglm',
                            'checkpoint':args.checkpoint,
                            'use_zero2_for_dp':args.use_zero2_for_dp,
                            'use_zero3_for_embed':args.use_zero3_for_embed,
                            'mixed_precision':args.mixed_precision,
                            'pipeline_type': args.pipeline_type,
                            'mixed_precision':args.mixed_precision}
    timecost_model_args_with_overlap = { 
                            'parameter_size': parameter_size,
                            'microbatch': microbatch,
                            'optimal_chunk_func': optimal_chunk_func,
                            'sequence_length': seq_len,
                            'hidden_size': hidden_size,
                            'forward_computation_time': forward_compute_time_per_layer,
                            'bct_fct_coe': 2,
                            'extra_overhead': 0,
                            'comm_coe_dict': comm_coe_dict,
                            'dp_overlap_coe': overlap_coe,
                            'bct_overlap_coe': overlap_coe,
                            'p2p_comm_coe_dict': p2p_comm_coe_dict,
                            'layer_num': layer_num,
                            'use_zero2_for_dp':args.use_zero2_for_dp,
                            'mixed_precision':args.mixed_precision}

    memcost_model_args_list = [memcost_model_args]
    timecost_model_args_list = [timecost_model_args_with_overlap]

    def pp_stage_divide_greedy(memcost_model_args, layer_num, pp_deg, bsz, strategies):
        assert(len(memcost_model_args)==len(layer_num))
        if pp_deg == 1:
            return [np.sum(layer_num)], None
        layer_type_num = len(layer_num)
        layer_min_memcost = []
        strategies = list(filter(lambda s: s[0] == pp_deg, strategies))
        if len(strategies)==0:
            return None, None
        for i in range(layer_type_num):
            memcosts = [MemoryCostModelDist(strategy, global_batch_size=bsz, **memcost_model_args[i]).get_memory_cost()['enc_total'] for strategy in strategies]
            layer_min_memcost.append(np.min(memcosts))
        other_cost = MemoryCostModelDist(strategies[0], global_batch_size=bsz, **memcost_model_args[0]).get_memory_cost()['other']
        #print(layer_min_memcost, other_cost)
        min_memcost_all_layers = []
        for i in range(layer_type_num):
            min_memcost_all_layers += [layer_min_memcost[i]]*layer_num[i]
        #print(min_memcost_all_layers)
        avg_mem_cost = (np.sum(min_memcost_all_layers)+np.sum(other_cost))/pp_deg
        #print('Avg memcost:', avg_mem_cost)

        pp_divide = [0]*pp_deg
        mem_cost_per_stage = other_cost.copy()
        idx = len(min_memcost_all_layers)-1
        for i in range(pp_deg-1,-1,-1):
            while True:
                if idx < 0:
                    break
                if i > 0 and avg_mem_cost - mem_cost_per_stage[i] < 0.5 * min_memcost_all_layers[idx]:
                    break
                else:
                    mem_cost_per_stage[i]+=min_memcost_all_layers[idx]
                    idx-=1
                    pp_divide[i]+=1
        # print(pp_divide)

        # Avoid too much memory cost on previous stages
        for i in range(pp_deg-1):
            left, right = int(np.sum(pp_divide[:i])), int(np.sum(pp_divide[:i+1]))
            mem_cost_cur_stage = np.sum(min_memcost_all_layers[left:right]) + other_cost[i]
            while mem_cost_cur_stage > avg_mem_cost * 1.3:
                pp_divide[i] -= 1
                pp_divide[i+1] += 1
                right -= 1
                mem_cost_cur_stage -= min_memcost_all_layers[right]

        # Avoid no layers on previous stages
        for i in range(pp_deg-1):
            while pp_divide[i] <= 0:
                pp_divide[i] += 1
                pp_divide[i+1] -= 1

        # Avoid no layers on last stage
        for i in range(pp_deg-1, 0, -1):
            while pp_divide[i] <= 0:
                pp_divide[i] += 1
                pp_divide[i-1] -= 1
        
        mem_cost_per_stage_adjusted = other_cost.copy()
        # print(pp_divide)
        # print(other_cost, avg_mem_cost)
        for i in range(pp_deg):
            left, right = int(np.sum(pp_divide[:i])), int(np.sum(pp_divide[:i+1]))
            mem_cost_per_stage_adjusted[i] +=  np.sum(min_memcost_all_layers[left:right])
        # print(mem_cost_per_stage,mem_cost_per_stage_adjusted)
        return pp_divide, mem_cost_per_stage_adjusted

    def get_pp_stages_for_all_bsz():
        bszs = list(range(8, 1024, 8))
        pp_stage_dict_for_bsz = dict()
        for bsz in bszs:
            # pp_stage_dict = dict()
            # pp_deg_list = sorted(list(set([s[0] for s in strategies])))
            # for pp_deg in pp_deg_list:
            #     pp_divide, mem_cost_per_stage = pp_stage_divide_greedy(memcost_model_args_list, layer_num_list, pp_deg, bsz, strategies)
            #     #print(bsz, pp_deg, pp_divide, mem_cost_per_stage)
            #     pp_stage_dict[pp_deg] = pp_divide
            # pp_stage_dict_for_bsz[bsz] = pp_stage_dict
            pp_stage_dict = {}
            for pp_deg in set([s[0] for s in strategies]):
                pp_stage_dict[pp_deg] = [layer_num // pp_deg] * pp_deg
            pp_stage_dict_for_bsz[bsz] = pp_stage_dict
        return pp_stage_dict_for_bsz

    search_history = dict()
    def search(max_mem):
        bsz_scale = 16
        bsz_scale = 16 if args.type == 'dp+tp' and bsz_scale < 16 else bsz_scale
        bsz_start = bsz_scale if args.search_from_min_bsz else estimate_bsz_start(bsz_scale)
        # bsz_start = 80
        print('Searching batch_size start from: %d, batch_size scale: %d'%(bsz_start, bsz_scale))
        print("----Searching with max memory %d MB----"%max_mem)
        # dp_on_model = DpOnModel_dist(strategies, 
        #                         MemoryCostModelDist, 
        #                         TimeCostModelDist_with_overlap, 
        #                         memcost_model_args,
        #                         timecost_model_args_with_overlap,
        #                         max_mem=max_mem,
        #                         search_history=search_history,
        #                         layer_num=layer_num,
        #                         comm_coe_dict=comm_coe_dict,
        #                         gpu_num=gpu_num)
        results = dict()
        max_throughput, optimal_bsz, max_bsz = -1, -1, -1
        for bsz in range(bsz_start, 1024, bsz_scale):

            pp_stage_dict = pp_stage_dict_for_bsz[bsz]
            print('bsz=%d'%bsz, pp_stage_dict)
            dp_on_model = DpOnModel_dist(strategies, 
                                    MemoryCostModelDist, 
                                    TimeCostModelDist_with_overlap, 
                                    memcost_model_args=memcost_model_args_list,
                                    timecost_model_args=timecost_model_args_list,
                                    max_mem=max_mem,
                                    layer_num =[layer_num],
                                    multi_layer_type = True,
                                    pp_stage_dict = pp_stage_dict,
                                    search_history=search_history,
                                    comm_coe_dict=comm_coe_dict,
                                    gpu_num=gpu_num,
                                    pipeline_type=args.pipeline_type)
            
            if args.settle_bsz > 0:
                bsz = args.settle_bsz
            print("****Testing with bsz=", bsz, "****")
            
            chunk_dict = check_optimal_chunks(args.gpu_num, strategies, optimal_chunk_func, bsz)
            print('Chunk_dict for bsz %d: '%bsz, chunk_dict)
            
            min_cost, min_res_list, min_pp_deg, mem_remain, mem_cost = dp_on_model.fit(bsz)
            throughput = bsz / min_cost
            print(f"[Optimal pp_deg={min_pp_deg}] Minimized timecost={min_cost} Memory remaining={mem_remain} Memory cost={mem_cost}")
            print(f"Max throughput={throughput} samples/s")
            print_strategies(min_res_list)
            results[bsz] = {'min_cost': min_cost, 'min_res_list': min_res_list, 'min_pp_deg': min_pp_deg, 
                            'mem_remain': mem_remain, 'mem_cost': mem_cost, 'throughput': throughput}
            if throughput > max_throughput:
                max_throughput = throughput
                optimal_bsz = bsz
            if min_pp_deg == -1 and min_res_list is None:
                break
            max_bsz = bsz
            if args.settle_bsz > 0:
                break

        print('\nFinal results of max memory %d MB:'%max_mem)
        re = results[optimal_bsz]
        print(f"Optimal bsz = {optimal_bsz} Max throughput={re['throughput']} samples/s")
        print(f"pp_deg={re['min_pp_deg']} Minimized timecost={re['min_cost']} Memory remaining={re['mem_remain']} Memory cost={re['mem_cost']}")
        print_strategies(re['min_res_list'])

        if re['min_pp_deg'] > 0 and re['min_res_list'] is not None:
            result_strategy = []
            if isinstance(re['min_res_list'],list):
                for l in re['min_res_list']:
                    result_strategy += l
            else:
                result_strategy = re['min_res_list']
            config = strategy2config(result_strategy)
            config['global_bsz'] = optimal_bsz
            config['chunks'] = max([int(optimal_chunk_func(optimal_bsz//s[2],s)) for s in result_strategy]) if config['pp_deg'] > 1 else 1
            config['pp_division'] = array2str(pp_stage_dict_for_bsz[optimal_bsz][config['pp_deg']])
            cpt = '_cpt' if args.checkpoint else ''
            if args.checkpoint:
                config['checkpoint'] = array2str([1 if 'cpt' in s[-1] and s[-1]['cpt'] else 0 for s in result_strategy])
            settle_bsz = '_bsz%d'%args.settle_bsz if args.settle_bsz > 0 else ''
            disable_sdp = '_sdpoff' if args.disable_sdp else ('_dpoff' if args.disable_dp else '')
            zero2 = '_zero2' if args.use_zero2_for_dp else ''
            bf16 = '_bf16' if args.mixed_precision else ''
            pp_type = '_1f1b' if args.pipeline_type == 'pipedream_flush' else ''
            config['pipeline_type'] = args.pipeline_type
            file_name = './configs/galvatron_config_%dgpus_%dhidden_%dlayers_%dG_%s%s%s%s%s%s%s.json'%(gpu_num,hidden_size,layer_num,max_mem//1024,args.type,cpt,pp_type,settle_bsz,disable_sdp,zero2,bf16)
            write_json_config(config, file_name)
            print('Already written optimized parallelism config into galvatron config file %s!'%(file_name))

        if max_bsz > -1 and max_bsz != optimal_bsz:
            re = results[max_bsz]
            print(f"\nMax bsz = {max_bsz} Max throughput={re['throughput']} samples/s")
            print(f"pp_deg={re['min_pp_deg']} Minimized timecost={re['min_cost']} Memory remaining={re['mem_remain']} Memory cost={re['mem_cost']}")
            print_strategies(re['min_res_list'])
        print("-----------------------------------------")


    # Check cost model
    def check_cost_model():
        bsz = 32
        for strategy in strategies:
            re = MemoryCostModelDist(strategy, global_batch_size=bsz, **memcost_model_args).get_memory_cost()
            print(form_strategy(strategy), re['enc_total'], re['other'][0], re['other'][-1], re['enc_total']*args.layer_num/strategy[0]+re['other'][0], re['enc_total']*args.layer_num/strategy[0]+re['other'][-1])
        print()
        for strategy in strategies:
            re = TimeCostModelDist_with_overlap(strategy, global_batch_size=bsz, **timecost_model_args_with_overlap).gen_result()
            print(form_strategy(strategy), re*args.layer_num)
        print()

    def estimate_bsz_start(scale):
        def estimate_strategy_max_bsz(s):
            dp_on_model = DpOnModel_dist(s, MemoryCostModelDist, TimeCostModelDist_with_overlap, 
                                    memcost_model_args, timecost_model_args_with_overlap,
                                    max_mem=max_mem, layer_num=layer_num, comm_coe_dict=comm_coe_dict, gpu_num=gpu_num)
            max_bsz = 0
            scale_ = 16 if s[0][0] == 1 and scale < 16 else scale
            for bsz in range(scale_, 1024, scale_):
                min_cost, min_res_list, min_pp_deg, mem_remain, mem_cost = dp_on_model.fit(bsz, False)
                if min_pp_deg == -1:
                    max_bsz = bsz - scale_
                    break
            return max_bsz
        bsz_start = estimate_bsz_start_16gpus(args.type,scale,estimate_strategy_max_bsz)
        return bsz_start

    check_cost_model()
    pp_stage_dict_for_bsz = get_pp_stages_for_all_bsz()
    # print(pp_stage_dict_for_bsz)
    mem_list = [8, 12, 16, 20]
    if args.memory_constraint > 0:
        mem_list = [args.memory_constraint]
    mem_list = [mem * 1024 for mem in mem_list]
    for max_mem in mem_list:
        search(max_mem)
        print()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--hidden_size", type=int, default=1280, help="Hidden size of transformer model",
    )
    parser.add_argument(
        "--layer_num", type=int, default=32, help="Number of layers"
    )
    parser.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    parser.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=32,
        help="Number of attention heads",
    )
    parser.add_argument(
        "--gpu_num", type=int, default=32, help="Number of GPUs",
    )
    parser.add_argument(
        "--memory_constraint", type=int, default=8, help="Memory constraint of Galvatron",
    )
    parser.add_argument(
        "--type", type=str, default='full', help="Galvatron parallelism optimization type.", choices=['full','dp+tp','dp+pp'],
    )
    parser.add_argument(
        "--search_from_min_bsz", type=int, default=0, help="If 0, start searching from a recommended bsz to accelerate optimization.",
    )
    # parser.add_argument(
    #     "--model_size", type=str, default='gpt-1.5b', help="Model size.", choices=['gpt-1.5b', 'gpt-2.7b', 'gpt-6.7b']
    # )
    parser.add_argument(
        "--checkpoint", type=int, default=0, help="Enable checkpoint or not."
    )
    parser.add_argument(
        "--settle_bsz", type=int, default=-1, help="Settle batch size or not."
    )
    parser.add_argument(
        "--disable_sdp", type=int, default=0, help="Whether to disable sdp."
    )
    parser.add_argument(
        "--disable_dp", type=int, default=0, help="Whether to disable dp."
    )
    parser.add_argument(
        "--disable_tp_consec", type=int, default=0, help="Whether to disable tp_consec."
    )
    parser.add_argument(
        "--use_zero2_for_dp", type=int, default=0, help="Whether to use_zero2_for_dp."
    )
    parser.add_argument(
        "--use_zero3_for_embed", type=int, default=0, help="Whether to use zero3 for embedding and cls layers."
    )
    parser.add_argument(
        "--mixed_precision", type=int, default=0, help="Whether to use mixed precision."
    )
    parser.add_argument(
        "--pipeline_type", type=str, default="gpipe", help="Galvatron pipeline type", choices=["gpipe","pipedream_flush"],
    )
    args = parser.parse_args()
    parallelism_optimization(args)


import torch
from torch import nn
from torch import Tensor, device
from typing import Tuple
import sys
from megatron.model.utils import init_method_normal, scaled_init_method_normal
from megatron import get_args
from megatron_layers import ParallelMLP, ParallelAttention
from megatron.model.enums import AttnMaskType, AttnType

class GPTMLP_tp(nn.Module):
    def __init__(self, config, tp_group = None):
        super().__init__()
        args=get_args()
        args.bias_gelu_fusion = True
        init_method = init_method_normal(args.init_method_std)
        scaled_init_method = scaled_init_method_normal(args.init_method_std, args.num_layers)
        self.tp_group = tp_group.group if tp_group is not None else None
        self.mlp = ParallelMLP(init_method, scaled_init_method, tp_group = self.tp_group)

    def forward(self, hidden_states):
        hidden_states, _ = self.mlp(hidden_states)
        return hidden_states


import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from torch.optim import Adam
from .dataloader import DataLoaderForGPT
import argparse
from tqdm import tqdm
import numpy as np
import random
import h5py
import time
import os
import sys
from ..utils import print_peak_memory
from transformers.models.gpt2.configuration_gpt2 import GPT2Config
from flash_attn.models.gpt import GPTLMHeadModel
from ..utils import print_peak_memory
from ..utils import read_json_config, write_json_config
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp.api import ShardingStrategy, MixedPrecision, BackwardPrefetch
from .gpt_config_utils import gpt_config, overwrite_configs_and_args

def set_seed():
    seed = 123
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

def model_forward(model, input_ids):
    lm_logits = model(input_ids=input_ids).logits
    shift_logits = lm_logits[..., :-1, :].contiguous()
    shift_labels = input_ids[..., 1:].contiguous()
    from flash_attn.losses.cross_entropy import CrossEntropyLoss
    loss_fn = CrossEntropyLoss()
    loss = loss_fn(shift_logits.view(-1, model.config.vocab_size), shift_labels.view(-1).long())
    return loss

def construct_fsdp_model(model, device):
    mixed_precision = {'fp32': torch.float, 'fp16': torch.float16, 'bf16': torch.bfloat16}[args.mixed_precision]
    mixed_precision_policy = MixedPrecision(
        param_dtype=mixed_precision, # Param precision
        reduce_dtype=mixed_precision, # Gradient communication precision
        buffer_dtype=mixed_precision, # Buffer precision
        cast_forward_inputs=True,
        cast_root_forward_inputs=True
    )
    backward_prefetch=None
    fsdp_args = dict(sharding_strategy = ShardingStrategy.NO_SHARD, 
                    mixed_precision=mixed_precision_policy, 
                    backward_prefetch=None,
                    auto_wrap_policy=None)

    model = model.to(device)
    model = FSDP(model, **fsdp_args)
    return model

def run_forward(args, layer_num):
    local_rank = args.local_rank
    rank = torch.distributed.get_rank()
    torch.cuda.set_device(local_rank)
    device = torch.device("cuda", local_rank)
    world_size = torch.distributed.get_world_size()

    config = gpt_config(args)
    overwrite_configs_and_args(config, args)
    args.num_hidden_layers = config.n_layer = layer_num
    print(config)

    print("Creating Dataloader...")
    dataset = DataLoaderForGPT(args)
    trainloader = DataLoader(dataset=dataset,
                            batch_size=args.train_batch_size,
                            shuffle=False)

    print("Creating Model...")
    model = GPTLMHeadModel(config, device='cpu')

    model = construct_fsdp_model(model, device)

    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start_iter, end_iter = 10, 20

    print("Profiling...")
    time_list = []
    for ep in range(args.epochs):
        for iter, batch in enumerate(trainloader):
            if iter == end_iter:
                avg_time = sum(time_list) / len(time_list) / args.train_batch_size
                print("Average forward computation time of %d GPT layers (hidden_size=%d) is: %.4f ms / bsz"%(args.num_hidden_layers, args.hidden_size, avg_time))
                return avg_time
            input_ids = batch.to(device)
            torch.cuda.synchronize()
            start.record()
            # model forward
            model_forward(model, input_ids)
            end.record()
            torch.cuda.synchronize()
            iter_time = start.elapsed_time(end)
            if iter >= start_iter:
                time_list.append(iter_time)

def profile(args):
    torch.distributed.init_process_group(backend="nccl")
    
    args.num_hidden_layers = 4
    time_2_layers = run_forward(args, args.num_hidden_layers)
    args.num_hidden_layers = 2
    time_1_layers = run_forward(args, args.num_hidden_layers)
    time_per_layer = (time_2_layers-time_1_layers)/2

    fwd_config_path = './configs/forward_profiling_config.json'
    config = read_json_config(fwd_config_path) if os.path.exists(fwd_config_path) else dict()
    fp16 = '_'+args.mixed_precision
    key = 'fwd_time%s_hidden_%d'%(fp16, args.hidden_size)
    config[key] = time_per_layer
    print('********************')
    print("Average forward computation time of GPT layer (hidden_size=%d) is: %.4f ms / layer / bsz"%(args.hidden_size, time_per_layer))
    write_json_config(config, fwd_config_path)
    print('Already written forward profiling config into env config file %s!\n'%(fwd_config_path))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--local-rank", type=int, default=-1, help="Local rank.",
    )
    parser.add_argument(
        '--gpu_id', type=int, default=0, help='Id of GPU to run.'
    )
    parser.add_argument(
        "--model_size", type=str, default='gpt-1.5b', help="Model size.", choices=['gpt-1.5b', 'gpt-2.7b', 'gpt-6.7b']
    )
    parser.add_argument(
        "--overwrite_config", type=int, default=0, help="Whether to overwrite model config"
    )
    parser.add_argument(
        "--train_batch_size", type=int, default=32, help="Training batch size"
    )
    parser.add_argument(
        "--hidden_size", type=int, default=768, help="Hidden size of transformer model",
    )
    parser.add_argument(
        "--num_hidden_layers", type=int, default=12, help="Number of layers"
    )
    parser.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=12,
        help="Number of attention heads",
    )
    parser.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    parser.add_argument(
        "--vocab_size", type=int, default=30522, help="Total number of vocab"
    )
    parser.add_argument(
        "--dropout_prob", type=float, default=0.1, help="Dropout rate."
    )
    parser.add_argument("--max_predictions_per_seq", type=int, default=20)
    parser.add_argument("-e", "--epochs", type=int,
                        default=10, help="Number of epochs")
    parser.add_argument("--lr", type=float, default=1e-4,
                        help="Learning rate of adam")
    parser.add_argument(
        "--adam_weight_decay", type=float, default=0.01, help="Weight_decay of adam"
    )

    parser.add_argument(
        "--check_loss", type=int, default=0, help="Whether to check model correctness."
    )
    parser.add_argument(
        "--profile", type=int, default=0, help="Whether to profile model GPU memory."
    )
    parser.add_argument(
        "--profile_type", type=str, default='allocated', help="Profile allocated memory or reserved memory.",
        choices = ['allocated', 'reserved'],
    )
    parser.add_argument(
        "--mixed_precision", type=str, default='bf16', help="Mixed precision option.", choices=['fp32', 'fp16', 'bf16'],
    )
    parser.add_argument(
        "--use_flash_attn", type=int, default=0, help="Whether to turn on flash attention."
    )

    args = parser.parse_args()
    set_seed()
    profile(args)


import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from torch.optim import Adam
from .dataloader import DataLoaderForCLIP
import argparse
from tqdm import tqdm
import numpy as np
import random
import h5py
import time
import os
import sys
from ..utils import print_peak_memory
from transformers import CLIPModel, CLIPConfig
from ..utils import print_peak_memory
from ..utils import read_json_config, write_json_config
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp.api import ShardingStrategy, MixedPrecision, BackwardPrefetch
from .config_utils import *

def set_seed():
    seed = 123
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

def model_forward(model, input_ids, attention_mask, pixel_values):
    loss = model(input_ids=input_ids, 
                 attention_mask=attention_mask, 
                 pixel_values=pixel_values, 
                 return_loss=True).loss
    return loss

def construct_fsdp_model(model, device):
    mixed_precision = {'fp32': torch.float, 'fp16': torch.float16, 'bf16': torch.bfloat16}[args.mixed_precision]
    mixed_precision_policy = MixedPrecision(
        param_dtype=mixed_precision, # Param precision
        reduce_dtype=mixed_precision, # Gradient communication precision
        buffer_dtype=mixed_precision, # Buffer precision
        cast_forward_inputs=True,
        cast_root_forward_inputs=True
    )
    backward_prefetch=None
    fsdp_args = dict(sharding_strategy = ShardingStrategy.NO_SHARD, 
                    mixed_precision=mixed_precision_policy, 
                    backward_prefetch=None,
                    auto_wrap_policy=None)

    model = model.to(device)
    model = FSDP(model, **fsdp_args)
    return model

def run_forward(args):
    local_rank = args.local_rank
    rank = torch.distributed.get_rank()
    torch.cuda.set_device(local_rank)
    device = torch.device("cuda", local_rank)
    world_size = torch.distributed.get_world_size()

    print("Creating Model...")
    config = CLIPConfig.from_pretrained(clip_configs(args.model_type))
    if args.overwrite_config:
        config.text_config.num_hidden_layers = args.num_hidden_layers_text
        config.vision_config.num_hidden_layers = args.num_hidden_layers_vision
    # print(obtain_main_config(config))
    
    model = CLIPModel(config)
    
    print("Creating Dataloader...")
    dataset = DataLoaderForCLIP(config)
    trainloader = DataLoader(dataset=dataset,
                            batch_size=args.train_batch_size,
                            shuffle=False)

    model = construct_fsdp_model(model, device)

    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start_iter, end_iter = 10, 20

    print("Profiling...")
    time_list = []
    for ep in range(args.epochs):
        for iter, batch in enumerate(trainloader):
            if iter == end_iter:
                avg_time = sum(time_list) / len(time_list) / args.train_batch_size
                return avg_time
            input_ids, attention_mask, pixel_values = [t.to(device) for t in batch]
            torch.cuda.synchronize()
            start.record()
            # model forward
            model_forward(model, input_ids, attention_mask, pixel_values)
            end.record()
            torch.cuda.synchronize()
            iter_time = start.elapsed_time(end)
            if iter >= start_iter:
                time_list.append(iter_time)

def profile(args):
    torch.distributed.init_process_group(backend="nccl")
    
    layer_types = ['vision', 'text']
    model_type = args.model_type
    type_num = len(layer_types)
    layernums = [8,16]
    
    def overwrite_layernums(args, layernum_list):
        args.num_hidden_layers_vision, args.num_hidden_layers_text = layernum_list[0], layernum_list[1]
    
    layernum_list = [layernums[0]] * type_num
    overwrite_layernums(args, layernum_list)
    time_base = run_forward(args)
    print("Average forward computation time of", layernum_list, layer_types, " layers (%s) is: %.4f ms / bsz"%(model_type, time_base))
    
    time_per_layer_list = []
    for i in range(type_num):
        layernum_list = [layernums[0]] * type_num
        layernum_list[i] = layernums[1]
        overwrite_layernums(args, layernum_list)
        time_layer_i = run_forward(args)
        print("Average forward computation time of", layernum_list, layer_types, " layers (%s) is: %.4f ms / bsz"%(model_type, time_layer_i))
        time_per_layer_list.append((time_layer_i-time_base)/(layernums[1]-layernums[0]))

    fwd_config_path = './configs/forward_profiling_config.json'
    config = read_json_config(fwd_config_path) if os.path.exists(fwd_config_path) else dict()
    fp16 = '_'+args.mixed_precision
    
    for i in range(type_num):
        key = 'fwd_time%s_%s_%s'%(fp16, model_type, layer_types[i])
        config[key] = time_per_layer_list[i]
        print('********************')
        print("Average forward computation time of %s layer (%s) is: %.4f ms / layer / bsz"%(layer_types[i], model_type, time_per_layer_list[i]))
    write_json_config(config, fwd_config_path)
    print('Already written forward profiling config into env config file %s!\n'%(fwd_config_path))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--local-rank", type=int, default=-1, help="Local rank.",
    )
    parser.add_argument(
        '--gpu_id', type=int, default=0, help='Id of GPU to run.'
    )
    parser.add_argument(
        "--model_type", type=str, default='vit-L-14', help="CLIP model type."
    )
    parser.add_argument(
        "--train_batch_size", type=int, default=32, help="Training batch size"
    )
    parser.add_argument(
        "--overwrite_config", type=int, default=1, help="Whether to overwrite model config"
    )
    parser.add_argument(
        "--hidden_size", type=int, default=768, help="Hidden size of transformer model",
    )
    parser.add_argument(
        "--num_hidden_layers_text", type=int, default=12, help="Number of layers"
    )
    parser.add_argument(
        "--num_hidden_layers_vision", type=int, default=12, help="Number of layers"
    )
    parser.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=12,
        help="Number of attention heads",
    )
    parser.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    parser.add_argument(
        "--vocab_size", type=int, default=30522, help="Total number of vocab"
    )
    parser.add_argument(
        "--dropout_prob", type=float, default=0.1, help="Dropout rate."
    )
    parser.add_argument("--max_predictions_per_seq", type=int, default=20)
    parser.add_argument("-e", "--epochs", type=int,
                        default=10, help="Number of epochs")
    parser.add_argument("--lr", type=float, default=1e-4,
                        help="Learning rate of adam")
    parser.add_argument(
        "--adam_weight_decay", type=float, default=0.01, help="Weight_decay of adam"
    )

    parser.add_argument(
        "--check_loss", type=int, default=0, help="Whether to check model correctness."
    )
    parser.add_argument(
        "--profile", type=int, default=0, help="Whether to profile model GPU memory."
    )
    parser.add_argument(
        "--profile_type", type=str, default='allocated', help="Profile allocated memory or reserved memory.",
        choices = ['allocated', 'reserved'],
    )
    parser.add_argument(
        "--mixed_precision", type=str, default='bf16', help="Mixed precision option.", choices=['fp32', 'fp16', 'bf16'],
    )
    parser.add_argument(
        "--flash_attn", type=int, default=0, help="Whether to turn on flash attention."
    )

    args = parser.parse_args()
    set_seed()
    profile(args)


import numpy as np

class MemoryCostModel:
    def __init__(self,
            strategy,
            global_batch_size = 8,
            parameter_size = 48,
            tp_activation_per_bsz_dict = {1:85, 2:47, 4:28, 8:18.5},
            other_model_states = 640, 
            other_activation_per_bsz = 320,
            pytorch_context_mem = 1024):
        self.strategy = strategy
        self.pp_size = self.strategy[0]
        self.tp_size = self.strategy[1]
        self.dp_size = self.strategy[2]
        self.parameter_size = parameter_size/self.tp_size
        self.model_states_size = 4 * self.parameter_size
        if 'fsdp' in self.strategy[-1].keys() and self.strategy[-1]['fsdp']:
            # fsdp_model_states memory is slightly larger than dp_model_states/dp_size
            # we add a small bias to ensure the predicted fsdp memory NOT smaller than real value
            # Actually, this bias barely affect search result.
            self.model_states_size  *= (1/self.dp_size + 0.025)
        self.bsz = global_batch_size/self.dp_size
        self.activation_size = tp_activation_per_bsz_dict[self.tp_size] * self.bsz
        self.total = self.model_states_size + self.activation_size
        self.other_memcost = other_model_states + other_activation_per_bsz * (global_batch_size/self.tp_size/self.dp_size)
        self.other_memcost += pytorch_context_mem

    def get_memory_cost(self):
        result = dict()
        result['parameter'] = self.parameter_size
        result['model_states'] = self.model_states_size
        result['activation'] = self.activation_size
        result['enc_total'] = self.total
        result['other'] = self.other_memcost
        return result

class TimeCostModel_with_overlap:
    def __init__(self,
            strategy,
            global_batch_size,
            parameter_size = 48,
            microbatch=True,
            optimal_chunk_func = None,
            sequence_length=512,
            hidden_size=1024,
            forward_computation_time=35 / 24,
            bct_fct_coe=2,
            extra_overhead=0,
            comm_coe_dict={},
            dp_overlap_coe=1.3,
            bct_overlap_coe=1.3,
            layer_type='enc'):
        self.s = strategy[:3]
        self.sl = sequence_length
        self.hs = hidden_size
        self.microbatch = microbatch
        self.pp_size = self.s[0]
        self.tp_size = self.s[1]
        self.dp_size = self.s[2]
        self.comm_coe_dict = comm_coe_dict[self.pp_size]
        if self.tp_size == 1 or self.dp_size == 1:
            self.dc = self.comm_coe_dict['%d'%self.dp_size]
            self.tc = self.comm_coe_dict['%d'%self.tp_size]
        else:
            # In this case, strategy[-1]['tp'] represents tp_consecutive_flag
            info = strategy[-1]
            assert 'tp' in info.keys() and info['tp'] in [0, 1]
            tp_consecutive_flag = info['tp']
            if tp_consecutive_flag:
                self.dc = self.comm_coe_dict['%d_0'%self.dp_size]
                self.tc = self.comm_coe_dict['%d_1'%self.tp_size]
            else:
                self.dc = self.comm_coe_dict['%d_1'%self.dp_size]
                self.tc = self.comm_coe_dict['%d_0'%self.tp_size]
        self.fsdp = False
        if 'fsdp' in strategy[-1].keys() and strategy[-1]['fsdp']:
            self.fsdp = True
        self.dp_overlap_coe = dp_overlap_coe
        self.dc_overlap = self.dc*dp_overlap_coe
        self.ps = parameter_size/self.tp_size
        self.bs = global_batch_size/self.dp_size 
        self.layer_type = layer_type
        assert(layer_type in ['enc', 'dec'])
        if microbatch:
            self.optimal_microbatch = optimal_chunk_func(self.bs, self.s)
            # print(self.optimal_microbatch)

        # Dummy layer_num, can be any multiple of 8.
        # We estimate the time cost of single layer by averaging the time of whole model to deal with pipeline parallel
        self.layer_num = 24 

        # forward & backward computation time of whole model (depending on dummy layer_num)
        self.fct = forward_computation_time * self.bs / self.tp_size * self.layer_num 
        self.bct = self.fct * bct_fct_coe
        self.bct_overlap_coe = bct_overlap_coe
        self.bct_overlap = self.bct*bct_overlap_coe
        self.eo = extra_overhead

        # dp & tp message size of whole model (depending on dummy layer_num)
        self.dp_message_size = (2*(self.dp_size-1)/self.dp_size*self.ps) * self.layer_num
        tp_comm_times = 4 if layer_type=='enc' else 6
        self.tp_message_size = 2*(self.tp_size-1)/self.tp_size*(self.bs*self.sl*self.hs*tp_comm_times*4/1024/1024) * self.layer_num

    def bct_dp_overlap(self, dp_message_size, bct):
        dp_overlap_time = dp_message_size * self.dc_overlap
        bct_overlap_time = bct * self.bct_overlap_coe
        if dp_overlap_time > bct_overlap_time:
            overlap_part = bct_overlap_time
            rest_part = (dp_message_size - bct_overlap_time / self.dc_overlap) * self.dc
            rest_dp_flag = True
        elif dp_overlap_time < bct_overlap_time:
            overlap_part = dp_overlap_time
            rest_part = (bct - dp_overlap_time / self.bct_overlap_coe) 
            rest_dp_flag = False
        else:
            overlap_part = bct_overlap_time
            rest_part = 0
            rest_dp_flag = False
        return overlap_part, rest_part, rest_dp_flag

    def pipe_with_microbatch(self, computation_overhead, communication_overhead):
        result = computation_overhead*(self.pp_size+self.optimal_microbatch-1)/(self.pp_size*self.optimal_microbatch)+communication_overhead
        return result

    def gen_result(self):
        if np.array_equal(self.s, [1,1,8]) or np.array_equal(self.s, [1,1,4]):
            overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct)
            result = self.fct + overlap_part + rest_part + self.eo
        elif np.array_equal(self.s, [1,2,4]) or np.array_equal(self.s, [1,2,2]):
            if self.layer_type == 'enc':
                overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct)
                result = self.fct + overlap_part + rest_part + self.tp_message_size*self.tc + self.eo
            elif self.layer_type == 'dec':
                overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct*2/3)
                result = self.fct + 1/3*self.bct + overlap_part + rest_part +self.tp_message_size*self.tc+self.eo
        elif np.array_equal(self.s, [1,4,2]) == True:
            if self.layer_type == 'enc':
                overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct*1/2)
                result = self.fct + 1/2*self.bct + overlap_part + rest_part + self.tp_message_size*self.tc + self.eo
            elif self.layer_type == 'dec':
                overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct*2/3)
                result = self.fct + 1/3*self.bct + overlap_part + rest_part + self.tp_message_size*self.tc + self.eo
        elif np.array_equal(self.s, [1,8,1]) or np.array_equal(self.s, [1,4,1]):
            result = self.fct + self.bct + self.tp_message_size*self.tc
        elif np.array_equal(self.s, [2,1,4]) or np.array_equal(self.s, [2,1,2]):
            bct_per_stage = self.bct / self.pp_size
            dp_message_size_per_stage = self.dp_message_size / self.pp_size
            overlap_part_per_stage, rest_part_per_stage, rest_dp_flag = self.bct_dp_overlap(dp_message_size_per_stage, bct_per_stage)
            if rest_dp_flag and not self.fsdp:
                overall_overhead = self.fct + overlap_part_per_stage * self.pp_size + rest_part_per_stage + self.eo
            else:
                overall_overhead = self.fct + (overlap_part_per_stage + rest_part_per_stage) * self.pp_size + self.eo
            if self.microbatch == False:
                result = overall_overhead
            else:
                computation_overhead = self.fct + self.bct
                communication_overhead = overall_overhead-computation_overhead
                result = self.pipe_with_microbatch(computation_overhead, communication_overhead)
        elif np.array_equal(self.s, [2,2,2]) == True:
            overlap_part, rest_part, _ = self.bct_dp_overlap(self.dp_message_size, self.bct*1/2)
            overall_overhead = self.fct + overlap_part + rest_part + self.bct*1/2 + self.tp_message_size*self.tc + self.eo
            if self.microbatch == False:
                result = overall_overhead
            else:
                computation_overhead = self.fct + self.bct + self.tp_message_size*self.tc
                communication_overhead = overall_overhead-computation_overhead
                result = self.pipe_with_microbatch(computation_overhead, communication_overhead)
        elif np.array_equal(self.s, [2,4,1]) or np.array_equal(self.s, [2,2,1]):
            if self.microbatch == False:
                result = self.fct + self.bct + self.tp_message_size*self.tc
            else:
                overall_overhead = self.fct + self.bct + self.tp_message_size*self.tc
                result = self.pipe_with_microbatch(overall_overhead, 0)
        elif np.array_equal(self.s, [4,1,2]) == True:
            # new version
            bct_per_stage = self.bct / self.pp_size
            dp_message_size_per_stage = self.dp_message_size / self.pp_size
            overlap_part_per_stage, rest_part_per_stage, rest_dp_flag = self.bct_dp_overlap(dp_message_size_per_stage, bct_per_stage)
            if rest_dp_flag and not self.fsdp:
                overall_overhead = self.fct + overlap_part_per_stage * self.pp_size + rest_part_per_stage + self.eo
            else:
                overall_overhead = self.fct + (overlap_part_per_stage + rest_part_per_stage) * self.pp_size + self.eo
            if self.microbatch == False:
                result = overall_overhead
            else:
                computation_overhead = self.fct + self.bct
                communication_overhead = overall_overhead-computation_overhead
                result = self.pipe_with_microbatch(computation_overhead, communication_overhead)
        elif np.array_equal(self.s, [4,2,1]) == True:
            if self.microbatch == False:
                result = self.fct + self.bct + self.tp_message_size*self.tc
            else:
                overall_overhead = self.fct + self.bct + self.tp_message_size*self.tc
                result = self.pipe_with_microbatch(overall_overhead, 0)
        elif np.array_equal(self.s, [8,1,1]) or np.array_equal(self.s, [4,1,1]):
            if self.microbatch == False:
                result = self.fct + self.bct
            else:
                overall_overhead = self.fct + self.bct
                result = self.pipe_with_microbatch(overall_overhead, 0)
        if self.fsdp:
            result = result + self.dp_message_size * 0.5 * self.dc
        coe = 0.0011
        result = result*coe
        result = result / self.layer_num
        return result

class TimeCostModel_alternative:
    def __init__(self,
            strategy,
            global_batch_size,
            parameter_size = 48,
            microbatch=True,
            optimal_chunk_func = None,
            sequence_length=512,
            hidden_size=1024,
            forward_computation_time=35 / 24,
            bct_fct_coe=2,
            extra_overhead=0,
            comm_coe_dict={},
            dp_overlap_coe=1.3,
            bct_overlap_coe=1.3,
            layer_type='enc',
            profiled_policy_cost={}):
        from .dp_utils import form_strategy
        key = form_strategy(strategy)
        self.bs = global_batch_size
        dic = profiled_policy_cost[key]
        self.bszs, self.vals = [], []
        for bsz, val in dic.items():
            if 'hidden%d'%hidden_size in bsz:
                self.bszs.append(int(bsz.split('_')[-1][3:]))
                self.vals.append(val)
        if self.bs not in self.bszs:
            vals = sorted(self.vals) if self.bs > self.bszs[-1] else self.vals
            self.func = np.poly1d(np.polyfit(self.bszs, vals, deg=1))

    def gen_result(self):
        if self.bs in self.bszs:
            result = self.vals[self.bszs.index(self.bs)]
        else:
            result = self.func(self.bs)
        result = result if result >= 0.0 else 0.0
        return result


import os
import sys
import torch
from torch import nn
import numpy as np
from ..utils import gen_groups_dist, modules_to_devices, wrap_modules_data_parallel, wrap_modules_relocation, wrap_data_parallel, wrap_modules_checkpoint
from ..utils import read_json_config, config2strategy, str2array
from ..pipeline import PipelineParallel, PipeSequential

def get_chunks(args):
    if args.chunks == -1:
        args.chunks = 1
        if args.pp_deg > 1:
            world_size = torch.distributed.get_world_size()
            max_dp_deg = world_size // args.pp_deg
            local_bsz = args.global_train_batch_size // max_dp_deg
            optimal_micro_bsz = np.ceil(local_bsz / 4)
            optimal_micro_bsz = 1 if optimal_micro_bsz == 0 else optimal_micro_bsz
            args.chunks = int(optimal_micro_bsz)
    return args.chunks

def overwrite_megatron_args(config, args):
    args.hidden_size = config.hidden_size
    args.num_layers = config.num_layers
    args.num_attention_heads = config.num_attention_heads
    assert config.hidden_size % config.num_attention_heads == 0
    args.kv_channels = config.hidden_size // config.num_attention_heads
    # args.ffn_hidden_size = config.inner_hidden_size
    args.max_position_embeddings = config.max_sequence_length
    # args.attention_dropout = config.attention_dropout
    # args.hidden_dropout = config.dropout
    args.use_cpu_initialization = True

def get_hybrid_parallel_configs(args):
    local_rank = args.local_rank
    world_size = torch.distributed.get_world_size()
    config_type = 'JSON' if args.galvatron_config_path not in [None,'None'] else 'PYTHON' if args.apply_strategy else 'GLOBAL'
    if local_rank == 0:
        print('======================== Galvatron Parallel Config =============================')
        print('Galvatron parallel config mode: [%s config mode]'%config_type)
    
    tp_sizes_enc, tp_consecutive_flags, dp_types_enc, pp_deg, bsz, chunks, checkpoint_flags_enc = list(), list(), list(), 1, 1, -1, list()
    if config_type == 'GLOBAL':
        pp_deg = args.pp_deg
        tp_sizes_enc = [args.global_tp_deg] * args.num_hidden_layers if args.global_tp_deg > 0 else [1]*args.num_hidden_layers
        tp_consecutive_flags = [args.global_tp_consec] * args.num_hidden_layers if args.global_tp_consec in [0, 1] else [1]*args.num_hidden_layers
        dp_types_enc = args.num_hidden_layers * [args.fsdp]
        checkpoint_flags_enc = [args.global_checkpoint] * args.num_hidden_layers
    else:
        if config_type == 'JSON':
            galvatron_config = read_json_config(args.galvatron_config_path)
            pp_deg, tp_sizes_enc, tp_consecutive_flags, dp_types_enc = config2strategy(galvatron_config)
            bsz, chunks = galvatron_config['global_bsz'], galvatron_config['chunks']
            if 'checkpoint' in galvatron_config.keys():
                checkpoint_flags_enc = str2array(galvatron_config['checkpoint'])
            else:
                checkpoint_flags_enc = [0] * len(tp_sizes_enc)
            config_source = 'Galvatron JSON config %s'%args.galvatron_config_path
        elif config_type == 'PYTHON':
            tp_sizes_enc, tp_consecutive_flags, dp_types_enc, pp_deg, bsz, chunks, checkpoint_flags_enc = apply_strategy()
            config_source = 'Galvatron PYTHON config'
        if local_rank == 0 and (args.num_hidden_layers != len(tp_sizes_enc) or args.chunks != chunks or args.global_train_batch_size != bsz):
            print('[Notice] The following hyper-parameters will be overwritten by Galvatron %s config:'%config_type)
            if args.global_train_batch_size != bsz:
                print('   global_batch_size =', bsz)
            if args.chunks != chunks:
                print('   chunks =', chunks)
            if args.num_hidden_layers != len(tp_sizes_enc):
                print('   num_hidden_layers =', len(tp_sizes_enc))
        args.global_train_batch_size = bsz
        args.chunks = chunks
        args.num_hidden_layers = len(tp_sizes_enc)

    avg_num_layers = args.num_hidden_layers // pp_deg
    pp_ranks_enc = []
    for i in range(pp_deg):
        pp_ranks_enc += [i] * avg_num_layers

    assert args.global_train_batch_size % (world_size//pp_deg) == 0, 'global_train_batch_size should be multiple of world_size//pp_deg!'
    if local_rank == 0:
        if config_type == 'GLOBAL':
            print('[GLOBAL config mode] Loaded global hybrid parallel strategy:')
            dp_type = 'sdp' if args.fsdp else 'dp'
            tp_deg, tp_consec = tp_sizes_enc[0], tp_consecutive_flags[0]
            dp_deg = world_size//args.global_tp_deg//args.pp_deg
            print('   global_batch_size: %d, chunks: %d'%(args.global_train_batch_size, get_chunks(args)))
            print('   pp_deg: %d, tp_deg: %d, %s_deg: %d, tp_consecutive_flag: %d, checkpoint_flag: %d'%(pp_deg, tp_deg, dp_type, dp_deg, tp_consec, args.global_checkpoint))
        else:
            print('[%s config mode] Loaded hybrid parallel config from %s:'%(config_type, config_source))
            print('   global_batch_size: %d, chunks: %d, pp_deg: %d'%(args.global_train_batch_size, args.chunks, pp_deg))
            print('   tp_sizes_enc:\t', tp_sizes_enc)
            print('   tp_consecutive_flags:', tp_consecutive_flags)
            print('   dp_types_enc:\t', dp_types_enc)
            print('   checkpoint_flags:\t', checkpoint_flags_enc)
        print('================================================================================')
    hybrid_parallel_configs =   {'pp_deg':pp_deg,
                                'tp_sizes_enc':tp_sizes_enc,
                                'tp_consecutive_flags':tp_consecutive_flags,
                                'dp_types_enc':dp_types_enc,
                                'pp_ranks_enc':pp_ranks_enc,
                                'checkpoint_flags_enc':checkpoint_flags_enc}
    return hybrid_parallel_configs

def construct_hybrid_parallel_model(model, model_config, training_args, hybrid_parallel_configs):
    chatglm_model, config, args, hp_configs = model, model_config, training_args, hybrid_parallel_configs
    pp_deg, tp_sizes_enc, tp_consecutive_flags, dp_types_enc, pp_ranks_enc, checkpoint_flags_enc = \
        hp_configs['pp_deg'], hp_configs['tp_sizes_enc'], hp_configs['tp_consecutive_flags'], hp_configs['dp_types_enc'], hp_configs['pp_ranks_enc'], hp_configs['checkpoint_flags_enc']
    assert config.num_layers == len(tp_sizes_enc)
    assert config.num_layers == len(dp_types_enc) 
    assert config.num_layers == len(pp_ranks_enc)
    assert config.num_layers == len(checkpoint_flags_enc)
    world_size = torch.distributed.get_world_size()
    for tp_size in tp_sizes_enc:
        assert tp_size <= world_size//pp_deg and (world_size//pp_deg) % tp_size == 0 and tp_size >= 1, 'Wrong tp_size!'
    for dp_type in dp_types_enc:
        assert dp_type == 0 or dp_type == 1 or dp_type is None, 'Wrong dp_type!'
    for pp_rank in pp_ranks_enc:
        assert pp_rank >= 0 and pp_rank <= pp_deg - 1, 'Wrong pp_rank!'

    # [Step 0] Construct sizes & groups
    # Construct tp_sizes / dp_types / pp_stages for whole model
    tp_sizes_whole_model = [1] + tp_sizes_enc + [1, 1]
    dp_types_whole_model = [0] + dp_types_enc + [0, 0]
    pp_ranks_whole_model = [0] + pp_ranks_enc + [pp_deg-1, pp_deg-1]
    tp_consecutive_whole_model = [1] + tp_consecutive_flags + [1, 1]
    # Construct pp_group / tp_groups / dp_groups / allgather_groups / slice_funcs
    pp_group, tp_groups_whole_model, dp_groups_whole_model, allgather_groups_whole_model, slice_funcs_whole_model = gen_groups_dist(tp_sizes_whole_model, pp_deg, tp_consecutive_whole_model, show_rank = 0)
    tp_groups_enc = tp_groups_whole_model[1:-2]

    # # [Step 1] Construct Tensor Parallel Block based on tp_groups
    from .ChatGLMForConditionalGeneration_tensor_parallel import ChatGLMSelfAttention_tp, ChatGLMGLU_tp, GLMBlock_tp
    
    for i in range(config.num_layers):
        layer = chatglm_model.transformer.layers[i]
        setattr(layer, 'attention', ChatGLMSelfAttention_tp(config, layer, tp_group=tp_groups_enc[i]))
        setattr(layer, 'mlp', ChatGLMGLU_tp(config, tp_group=tp_groups_enc[i]))
        chatglm_model.transformer.layers[i] = GLMBlock_tp(config, layer)

    # [Step 2] Construct Sequantial modules
    from .ChatGLMForConditionalGeneration_pipeline import ChatGLMEmbeddings_, ChatGLMlayer_, ChatGLMNorm_, ChatGLMLMHead_
    model = PipeSequential()
    model.add_module('embeddings', ChatGLMEmbeddings_(config, chatglm_model))
    for i in range(config.num_layers):
        model.add_module('encoder_%d'%i, ChatGLMlayer_(config, chatglm_model, i))
    model.add_module('norm', ChatGLMNorm_(config, chatglm_model))
    model.add_module('cls', ChatGLMLMHead_(config, chatglm_model))

    # [Step 3] Wrap Relocation modules if necessary
    model = wrap_modules_relocation(model, allgather_groups_whole_model, slice_funcs_whole_model)

    # [Step 4] Construct Pipeline Module and place the layers on corresponding devices
    chunks = get_chunks(args)
    seq_len, hidden_size = args.seq_length, args.hidden_size
    mixed_precision = {'fp32': torch.float, 'fp16': torch.float16, 'bf16': torch.bfloat16}[args.mixed_precision]
    layer_output_tensor_shapes = [None] + [[[-1,seq_len,hidden_size], [-1,2,seq_len], [-1, 1, seq_len, seq_len]]] * config.num_layers + [None, None]
    layer_output_tensor_dtypes = [None] + [[mixed_precision, torch.long, torch.bool]] * config.num_layers + [None, None]

    layer_dp_sizes = [world_size // pp_deg // tp_size for tp_size in tp_sizes_whole_model]
    hp_model = PipelineParallel(
                model = model,
                model_ranks = pp_ranks_whole_model,
                layer_output_tensor_shapes = layer_output_tensor_shapes,
                layer_output_tensor_dtypes = layer_output_tensor_dtypes,
                layer_dp_sizes = layer_dp_sizes,
                chunks=args.chunks,
                process_group = pp_group.ranks,
                nproc_per_node=8,
                info=False)

    # [Step 5] Wrap Data Parallel modules based on dp_types & dp_groups
    module_types = ['embed'] + ['chatglm_enc']*config.num_layers + ['norm', 'cls']
    hp_model.wrap_pipeline_modules_data_parallel(dp_types_whole_model, dp_groups_whole_model, module_types=module_types, mixed_precision=mixed_precision)
    
    checkpoint_flags_whole_model = [0] + checkpoint_flags_enc + [0, 0]
    hp_model.wrap_pipeline_modules_checkpoint(checkpoint_flags_whole_model)
    
    return hp_model


def apply_strategy():
    ### Example strategy
    # For Bert Large
    pp_deg = 2
    tp_sizes_enc = [1]*24
    tp_consecutive_flags = [1]*24
    dp_types_enc = ([1]*6+[0]*6)*2
    global_bsz = 16
    chunks = 2
    checkpoint = [0]*24

    return tp_sizes_enc, tp_consecutive_flags, dp_types_enc, pp_deg, global_bsz, chunks, checkpoint


import torch
from torch import nn
from torch.utils.data import DataLoader
from .dataloader import DataLoaderForChatGLM
from transformers import AutoModel, AutoConfig
import numpy as np
import random
import os
import sys
from ..utils import read_json_config, write_json_config
from megatron.initialize import initialize_megatron
from megatron import get_args
from .hybrid_parallel_model_dist import overwrite_megatron_args
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp.api import ShardingStrategy, MixedPrecision, BackwardPrefetch

torch.backends.cudnn.allow_tf32 = True
torch.backends.cuda.matmul.allow_tf32 = True

def set_seed():
    seed = 123
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)

def model_forward(model, input_ids, labels):
    lm_logits = model(input_ids).logits
    loss_fct = nn.CrossEntropyLoss()

    loss = None
    lm_logits = lm_logits.to(torch.float32)

    # Shift so that tokens < n predict n
    shift_logits = lm_logits[..., :-1, :].contiguous()
    shift_labels = labels[..., 1:].contiguous()
    # Flatten the tokens
    loss_fct = nn.CrossEntropyLoss(ignore_index=-100)
    loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

    return loss

def construct_fsdp_model(model, device):
    # rriginal model has all torch.half params_dtype except layernorms
    glm_model = model.transformer
    glm_model.final_layernorm = nn.LayerNorm(glm_model.hidden_size, eps=glm_model.layernorm_epsilon, dtype=torch.half)
    for block in glm_model.layers:
        block.input_layernorm = nn.LayerNorm(glm_model.hidden_size, eps=glm_model.layernorm_epsilon, dtype=torch.half)
        block.post_attention_layernorm = nn.LayerNorm(glm_model.hidden_size, eps=glm_model.layernorm_epsilon, dtype=torch.half)

    mixed_precision = {'fp32': torch.float, 'fp16': torch.float16, 'bf16': torch.bfloat16}[args.mixed_precision]
    mixed_precision_policy = MixedPrecision(
        param_dtype=mixed_precision, # Param precision
        reduce_dtype=mixed_precision, # Gradient communication precision
        buffer_dtype=mixed_precision, # Buffer precision
        cast_forward_inputs=True,
        cast_root_forward_inputs=True
    )
    fsdp_args = dict(sharding_strategy = ShardingStrategy.NO_SHARD, 
                    mixed_precision=mixed_precision_policy, 
                    backward_prefetch=None,
                    auto_wrap_policy=None)

    model = model.to(device)
    model = FSDP(model, **fsdp_args)
    return model

def run_forward(args):
    local_rank = args.local_rank
    cuda_condition = torch.cuda.is_available()
    device = torch.device("cuda:%d"%args.gpu_id if cuda_condition else "cpu")
    rank = args.gpu_id

    config = AutoConfig.from_pretrained("THUDM/chatglm-6b", trust_remote_code=True)
    overwrite_config = {'hidden_size': args.hidden_size,
                        'inner_hidden_size': args.hidden_size*4,
                        'max_sequence_length': args.seq_length,
                        'num_attention_heads': args.num_attention_heads,
                        'num_layers': args.num_hidden_layers,
                        'vocab_size': args.vocab_size}
    
    for key, val in overwrite_config.items():
        setattr(config, key, val)
    
    overwrite_config = {'use_flash_attn': args.use_flash_attn}
    
    for key, val in overwrite_config.items():
        setattr(config, key, val)
    
    overwrite_megatron_args(config, args)

    print("Creating Dataloader...")
    dataset = DataLoaderForChatGLM(args, config)
    trainloader = DataLoader(dataset=dataset,
                            batch_size=args.train_batch_size,
                            shuffle=False)

    print("Creating Model...")
    
    chatglm_model = AutoModel.from_config(config, trust_remote_code=True)
    model = construct_fsdp_model(chatglm_model, device)
    
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start_iter, end_iter = 10, 20
    
    print("Profiling...")
    time_list = []
    for ep in range(args.epochs):
        for iter, batch in enumerate(trainloader):
            if iter == end_iter:
                avg_time = sum(time_list) / len(time_list) / args.train_batch_size
                print("Average forward computation time of %d ChatGLM layers (hidden_size=%d) is: %.4f ms / bsz"%(args.num_hidden_layers, args.hidden_size, avg_time))
                return avg_time
            input_ids, labels = [tensor.to(device) for tensor in batch]
            torch.cuda.synchronize()
            start.record()
            model_forward(model, input_ids, labels)
            end.record()
            torch.cuda.synchronize()
            iter_time = start.elapsed_time(end)
            if iter >= start_iter:
                time_list.append(iter_time)


def profile(args):
    args.num_hidden_layers = 8
    time_8_layers = run_forward(args)
    args.num_hidden_layers = 4
    time_4_layers = run_forward(args)
    time_per_layer = (time_8_layers-time_4_layers)/4

    fwd_config_path = './configs/forward_profiling_config.json'
    config = read_json_config(fwd_config_path) if os.path.exists(fwd_config_path) else dict()
    mp = '_'+args.mixed_precision
    key = 'fwd_time%s_hidden_%d'%(mp, args.hidden_size)
    config[key] = time_per_layer
    print('********************')
    print("Average forward computation time of ChatGLM layer (hidden_size=%d) is: %.4f ms / layer / bsz"%(args.hidden_size, time_per_layer))
    write_json_config(config, fwd_config_path)
    print('Already written forward profiling config into env config file %s!\n'%(fwd_config_path))


def add_arguments(parser):
    group = parser.add_argument_group(title='our arguments')

    group.add_argument(
        '--gpu_id', type=int, default=0, help='Id of GPU to run.'
    )
    group.add_argument(
        "--train_batch_size", type=int, default=32, help="Global training batch size"
    )
    group.add_argument(
        "--hidden_size", type=int, default=1280, help="Hidden size of transformer model",
    )
    group.add_argument(
        "--num_hidden_layers", type=int, default=12, help="Number of layers"
    )
    group.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=20,
        help="Number of attention heads",
    )
    group.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    group.add_argument(
        "--vocab_size", type=int, default=51865, help="Total number of vocab"
    )
    group.add_argument(
        "--dropout_prob", type=float, default=0.1, help="Dropout rate."
    )
    group.add_argument("-e", "--epochs", type=int,
                        default=10, help="Number of epochs")
    parser.add_argument(
        "--mixed_precision", type=str, default='bf16', help="Mixed precision option.", choices=['fp32', 'fp16', 'bf16'],
    )
    return parser

if __name__ == '__main__':
    initialize_megatron(extra_args_provider=add_arguments)
    args = get_args()
    set_seed()
    profile(args)



import torch
import torch.nn as nn
import torch.nn.functional as F
import sys
from megatron.model.utils import init_method_normal, scaled_init_method_normal
from megatron import get_args
from megatron_layers import ParallelMLP, ParallelAttention
from megatron.model.enums import AttnMaskType, AttnType
from megatron.core import tensor_parallel

try:
    from einops import rearrange
except ImportError:
    rearrange = None

def rotate_half(x):
    x1, x2 = x[..., :x.shape[-1] // 2], x[..., x.shape[-1] // 2:]
    return torch.cat((-x2, x1), dim=x1.ndim - 1)  # dim=-1 triggers a bug in earlier torch versions

@torch.jit.script
def apply_rotary_pos_emb_index(q, k, cos, sin, position_id):
    # position_id: [sq, b], q, k: [sq, b, np, hn], cos: [sq, 1, hn] -> [sq, b, 1, hn]
    cos, sin = F.embedding(position_id, cos.squeeze(1)).unsqueeze(2), \
        F.embedding(position_id, sin.squeeze(1)).unsqueeze(2)
    q, k = (q * cos) + (rotate_half(q) * sin), (k * cos) + (rotate_half(k) * sin)
    return q, k

# implementation in ChatGLM-6B huggingface repo
# def attention_fn(
#         self,
#         query_layer,
#         key_layer,
#         value_layer,
#         attention_mask,
#         hidden_size_per_partition,
#         layer_id,
#         layer_past=None,
#         scaling_attention_score=True,
#         use_cache=False,
# ):
#     if layer_past is not None:
#         past_key, past_value = layer_past[0], layer_past[1]
#         key_layer = torch.cat((past_key, key_layer), dim=0)
#         value_layer = torch.cat((past_value, value_layer), dim=0)

#     # seqlen, batch, num_attention_heads, hidden_size_per_attention_head
#     seq_len, b, nh, hidden_size = key_layer.shape

#     if use_cache:
#         present = (key_layer, value_layer)
#     else:
#         present = None

#     query_key_layer_scaling_coeff = float(layer_id + 1)
#     if scaling_attention_score:
#         query_layer = query_layer / (math.sqrt(hidden_size) * query_key_layer_scaling_coeff)

#     # ===================================
#     # Raw attention scores. [b, np, s, s]
#     # ===================================

#     # [b, np, sq, sk]
#     output_size = (query_layer.size(1), query_layer.size(2), query_layer.size(0), key_layer.size(0))

#     # [sq, b, np, hn] -> [sq, b * np, hn]
#     query_layer = query_layer.view(output_size[2], output_size[0] * output_size[1], -1)
#     # [sk, b, np, hn] -> [sk, b * np, hn]
#     key_layer = key_layer.view(output_size[3], output_size[0] * output_size[1], -1)

#     matmul_result = torch.zeros(
#         1, 1, 1,
#         dtype=query_layer.dtype,
#         device=query_layer.device,
#     )

#     matmul_result = torch.baddbmm(
#         matmul_result,
#         query_layer.transpose(0, 1),  # [b * np, sq, hn]
#         key_layer.transpose(0, 1).transpose(1, 2),  # [b * np, hn, sk]
#         beta=0.0,
#         alpha=1.0,
#     )

#     # change view to [b, np, sq, sk]
#     attention_scores = matmul_result.view(*output_size)

#     if self.scale_mask_softmax:
#         self.scale_mask_softmax.scale = query_key_layer_scaling_coeff
#         attention_probs = self.scale_mask_softmax(attention_scores, attention_mask.contiguous())
#     else:
#         if not (attention_mask == 0).all():
#             # if auto-regressive, skip
#             attention_scores.masked_fill_(attention_mask, -10000.0)
#         dtype = attention_scores.dtype
#         attention_scores = attention_scores.float()
#         attention_scores = attention_scores * query_key_layer_scaling_coeff

#         attention_probs = F.softmax(attention_scores, dim=-1)

#         attention_probs = attention_probs.type(dtype)

#     # =========================
#     # Context layer. [sq, b, hp]
#     # =========================

#     # value_layer -> context layer.
#     # [sk, b, np, hn] --> [b, np, sq, hn]

#     # context layer shape: [b, np, sq, hn]
#     output_size = (value_layer.size(1), value_layer.size(2), query_layer.size(0), value_layer.size(3))

#     # change view [sk, b * np, hn]
#     value_layer = value_layer.view(value_layer.size(0), output_size[0] * output_size[1], -1)

#     # change view [b * np, sq, sk]
#     attention_probs = attention_probs.view(output_size[0] * output_size[1], output_size[2], -1)

#     # matmul: [b * np, sq, hn]
#     context_layer = torch.bmm(attention_probs, value_layer.transpose(0, 1))

#     # change view [b, np, sq, hn]
#     context_layer = context_layer.view(*output_size)

#     # [b, np, sq, hn] --> [sq, b, np, hn]
#     context_layer = context_layer.permute(2, 0, 1, 3).contiguous()

#     # [sq, b, np, hn] --> [sq, b, hp]
#     new_context_layer_shape = context_layer.size()[:-2] + (hidden_size_per_partition,)
#     context_layer = context_layer.view(*new_context_layer_shape)

#     outputs = (context_layer, present, attention_probs)

#     return outputs

class ChatGLMSelfAttention_tp(nn.Module):
    def __init__(self, config, glm_block, tp_group=None):
        super().__init__()
        args=get_args()
        init_method = init_method_normal(args.init_method_std)
        scaled_init_method = scaled_init_method_normal(args.init_method_std, args.num_layers)
        self.tp_group = tp_group.group if tp_group is not None else None
        self.config = config
        self.position_encoding_2d = glm_block.attention.position_encoding_2d
        self.rotary_emb = glm_block.attention.rotary_emb

        self.attention = ParallelAttention(init_method,
                                        scaled_init_method,
                                        attention_type=AttnType.self_attn,
                                        attn_mask_type=AttnMaskType.causal, # GLM attn mask is non-causal + causal
                                        tp_group=self.tp_group)
    
    def forward(self, hidden_states, position_ids, attention_mask, layer_id, layer_past=None, use_cache=False, output_attentions=False):
        """
        hidden_states: [seq_len, batch, hidden_size]
        attention_mask: [(1, 1), seq_len, seq_len]
        """

        # [seq_len, batch, 3 * hidden_size]
        mixed_raw_layer, _ = self.attention.query_key_value(hidden_states) # zsh comment: Megatron impl has no bias but ChatGLM impl has

        # [seq_len, batch, 3 * hidden_size] --> [seq_len, batch, num_attention_heads, 3 * hidden_size_per_attention_head]
        new_tensor_shape = mixed_raw_layer.size()[:-1] + (
            self.attention.num_attention_heads_per_partition,
            3 * self.attention.hidden_size_per_attention_head,
        )
        mixed_raw_layer = mixed_raw_layer.view(*new_tensor_shape)

        # [seq_len, batch, num_attention_heads, hidden_size_per_attention_head]
        (query_layer, key_layer, value_layer) = tensor_parallel.split_tensor_along_last_dim(mixed_raw_layer, 3)

        if self.position_encoding_2d:
            q1, q2 = query_layer.chunk(2, dim=(query_layer.ndim - 1))
            k1, k2 = key_layer.chunk(2, dim=(key_layer.ndim - 1))
            cos, sin = self.rotary_emb(q1, seq_len=position_ids.max() + 1)
            position_ids, block_position_ids = position_ids[:, 0, :].transpose(0, 1).contiguous(), \
                position_ids[:, 1, :].transpose(0, 1).contiguous()
            q1, k1 = apply_rotary_pos_emb_index(q1, k1, cos, sin, position_ids)
            q2, k2 = apply_rotary_pos_emb_index(q2, k2, cos, sin, block_position_ids)
            query_layer = torch.concat([q1, q2], dim=(q1.ndim - 1))
            key_layer = torch.concat([k1, k2], dim=(k1.ndim - 1))
        else:
            position_ids = position_ids.transpose(0, 1)
            cos, sin = self.rotary_emb(value_layer, seq_len=position_ids.max() + 1)
            # [seq_len, batch, num_attention_heads, hidden_size_per_attention_head]
            query_layer, key_layer = apply_rotary_pos_emb_index(query_layer, key_layer, cos, sin, position_ids)

        # [seq_len, batch, hidden_size]
        
        # implementation in ChatGLM-6B huggingface repo
        # context_layer = attention_fn(
        #     self=self,
        #     query_layer=query_layer,
        #     key_layer=key_layer,
        #     value_layer=value_layer,
        #     attention_mask=attention_mask,
        #     hidden_size_per_partition=self.hidden_size_per_partition,
        #     layer_id=layer_id
        # )

        # implementation in megatron
        if not self.attention.use_flash_attn:
            if self.attention.checkpoint_core_attention:
                context_layer = self.attention._checkpointed_attention_forward(
                    query_layer, key_layer, value_layer, attention_mask)
            else:
                context_layer = self.attention.core_attention(
                    query_layer, key_layer, value_layer, attention_mask)
        else:
            q, k, v = [rearrange(x, 's b ... -> b s ...').contiguous()
                       for x in (query_layer, key_layer, value_layer)]
            if not self.attention.sequence_parallel:
                with tensor_parallel.get_cuda_rng_tracker().fork():
                    context_layer = self.attention.core_attention_flash(q, k, v)
            else:
                context_layer = self.attention.core_attention_flash(q, k, v)
            context_layer = rearrange(context_layer, 'b s h d -> s b (h d)').contiguous()
        

        output, bias = self.attention.dense(context_layer)
        output = output + bias # zsh add to fit ChatGLM

        outputs = (output, None, None)

        return outputs  # output, present, attention_probs


class ChatGLMGLU_tp(nn.Module):
    def __init__(self, config, tp_group=None):
        super().__init__()
        args = get_args()
        init_method = init_method_normal(args.init_method_std)
        scaled_init_method = scaled_init_method_normal(args.init_method_std, args.num_layers)
        self.tp_group = tp_group.group if tp_group is not None else None
        # should set args.openai_gelu = True
        args.openai_gelu = True
        self.mlp = ParallelMLP(init_method, scaled_init_method, tp_group=self.tp_group)

    def forward(self, hidden_states):
        hidden_states, bias = self.mlp(hidden_states)
        return hidden_states + bias


class GLMBlock_tp(nn.Module):
    def __init__(self, config, glm_block):
        super().__init__()
        self.config = config
        self.glm_block = glm_block
    
    def forward(self, hidden_states, position_ids, attention_mask, layer_id):
        hidden_states = hidden_states.permute(1, 0, 2)
        outputs = self.glm_block(hidden_states, position_ids, attention_mask, layer_id)
        outputs = (outputs[0].permute(1, 0, 2),)
        return outputs


# Copyright (c) 2023, Tri Dao.

import json
import math
import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import Union

import torch
import torch.nn.functional as F
from transformers import GPT2Config, LlamaConfig

def config_from_meta_checkpoint(checkpoint_path: Union[str, os.PathLike], model_name: str) -> LlamaConfig:
    """Load a LlamaConfig from a checkpoint path."""
    with open(Path(checkpoint_path) / model_name / 'params.json') as f:
        params = json.load(f)
    config = LlamaConfig(hidden_size=params['dim'], intermediate_size=None,
                         num_attention_heads=params['n_heads'],
                         num_hidden_layers=params['n_layers'],
                         rms_norm_eps=params['norm_eps'])
    return config


def config_from_hf_checkpoint(checkpoint_path: Union[str, os.PathLike], model_name: str) -> LlamaConfig:
    return LlamaConfig.from_pretrained(Path(checkpoint_path) / f'{model_name}-hf' / "config.json")


def config_from_checkpoint(
    checkpoint_path: Union[str, os.PathLike], model_name: str, checkpoint_format="meta"
) -> LlamaConfig:
    if checkpoint_format == "meta":
        return config_from_meta_checkpoint(checkpoint_path, model_name)
    else:
        return config_from_hf_checkpoint(checkpoint_path, model_name)


def llama_config_to_gpt2_config(llama_config: LlamaConfig) -> GPT2Config:
    return GPT2Config(
        vocab_size=llama_config.vocab_size,
        n_positions=llama_config.max_position_embeddings,
        n_embd=llama_config.hidden_size,
        n_layer=llama_config.num_hidden_layers,
        n_head=llama_config.num_attention_heads,
        n_inner=llama_config.intermediate_size,
        activation_function='swiglu',  # Hardcode since HF calls it 'silu'
        # Llama doesn't have dropout, idk if it's because they only release the inference code
        resid_pdrop=0.0,
        embd_pdrop=0.0,
        attn_pdrop=0.0,
        layer_norm_epsilon=llama_config.rms_norm_eps,
        initializer_range=llama_config.initializer_range,
        bos_token_id=llama_config.bos_token_id,
        eos_token_id=llama_config.eos_token_id,
        # These are new arguments not in the original GPT2Config
        pad_token_id=llama_config.pad_token_id,  # Idk if this does anything
        rms_norm=True,
        rotary_emb_fraction=1.0,
        rotary_emb_interleaved=True,
        tie_word_embeddings=False,
        qkv_proj_bias=False,
        out_proj_bias=False,
        mlp_fc1_bias=False,
        mlp_fc2_bias=False,
    )
    
def overwrite_configs_and_args(config, args):
    overwrite_config = {'use_cache': False,
                        'use_flash_attn': args.use_flash_attn,
                        'fused_bias_fc': True,
                        'sequence_parallel': False}
    for key, val in overwrite_config.items():
        setattr(config, key, val)
    
    if args.overwrite_config:
        overwrite_config = {'hidden_size': args.hidden_size,
                            'max_position_embeddings': args.seq_length,
                            'num_hidden_layers': args.num_hidden_layers,
                            'vocab_size': args.vocab_size}
        for key, val in overwrite_config.items():
            setattr(config, key, val)
    else:
        args.hidden_size = config.hidden_size
        args.seq_length = config.max_position_embeddings
        args.max_position_embeddings = config.max_position_embeddings
        args.num_hidden_layers = config.num_hidden_layers
        args.vocab_size = config.vocab_size



from .memory_utils import *
from .group_comm_utils import *
from .group_comm_utils_dist import gen_groups_dist
from .parallel_utils import *
from .allgather_utils import split_to_group, gather_from_group
from .dp_utils import DpOnModel, print_strategies, form_strategy, estimate_bsz_start_8gpus, estimate_bsz_start_4gpus
from .dp_utils_dist import DpOnModel_dist, estimate_bsz_start_16gpus, estimate_bsz_start_gpunum
from .cost_model import *
from .cost_model_dist import *
from .config_utils import *
from .strategies_utils import *
from .init_utils import *


import torch
from torch.utils.data import Dataset
import h5py
import numpy as np

class DataLoaderForLlama(Dataset):
    def __init__(self,args):
        self.vocab_size = args.vocab_size
        self.sentence_length = args.seq_length
        self.dataset_size = 2560 * 16
        self.data_length = np.random.randint(1,self.sentence_length+1,(self.dataset_size,))

        self.input_ids = []
        # self.attention_mask = []
        for i in range(self.dataset_size):
            sentence = np.random.randint(0,self.vocab_size,(self.sentence_length,))
            sentence[self.data_length[i]:] = 0
            mask = np.ones((self.sentence_length,))
            mask[self.data_length[i]:] = 0
            self.input_ids.append(sentence)
            # self.attention_mask.append(sentence)
        
        self.input_ids = np.array(self.input_ids)
        # self.attention_mask = np.array(self.attention_mask)

    def __len__(self):
        return self.dataset_size

    def __getitem__(self, idx):
        if idx >= self.dataset_size:
            raise IndexError
        input_ids = torch.LongTensor(self.input_ids[idx])
        # attention_mask = torch.LongTensor(self.attention_mask[idx])
        # return input_ids, attention_mask
        return input_ids


import torch
import torch.nn as nn
from torch.nn.utils import skip_init

class ChatGLMEmbeddings_(nn.Module):
    def __init__(self, config, model):
        super().__init__()
        self.config = config
        model = model.transformer
        self.word_embeddings = model.word_embeddings
        self.get_masks = model.get_masks
        self.get_position_ids = model.get_position_ids
        
        # attrs = ['word_embeddings', 'get_masks, 'get_position_ids']
        # for key in attrs:
        #     setattr(self, key, getattr(model, key))
    
    def forward(self, input_ids, position_ids=None, attention_mask=None):
        inputs_embeds = self.word_embeddings(input_ids)

        if attention_mask is None:
            attention_mask = self.get_masks(
                input_ids,
                device=input_ids.device
            )

        if position_ids is None:
            MASK, gMASK = self.config.mask_token_id, self.config.gmask_token_id
            seqs = input_ids.tolist()

            mask_positions, use_gmasks = [], []
            for seq in seqs:
                mask_token = gMASK if gMASK in seq else MASK
                use_gmask = mask_token == gMASK
                mask_positions.append(seq.index(mask_token))
                use_gmasks.append(use_gmask)

            position_ids = self.get_position_ids(
                input_ids,
                mask_positions=mask_positions,
                device=input_ids.device,
                use_gmasks=use_gmasks
            )
        
        # comment this line to remain the shape: [batch, seq_len, hidden_size]
        # hidden_states = inputs_embeds.transpose(0, 1)
        hidden_states = inputs_embeds

        if attention_mask is None:
            attention_mask = torch.zeros(1, 1, device=input_ids.device).bool()
        else:
            attention_mask = attention_mask.to(hidden_states.device)
        
        return hidden_states, position_ids, attention_mask


class ChatGLMlayer_(nn.Module):
    def __init__(self, config, model, layer_idx):
        super().__init__()
        self.config = config
        self.layer_idx = layer_idx
        self.layer = model.transformer.layers[layer_idx]
    
    def forward(self, hidden_states, position_ids, attention_mask):
        layer_ret = self.layer(
            hidden_states=hidden_states,
            position_ids=position_ids,
            attention_mask=attention_mask,
            layer_id=torch.tensor(self.layer_idx)
        )
        hidden_states = layer_ret[0]

        return hidden_states, position_ids, attention_mask


class ChatGLMNorm_(nn.Module):
    def __init__(self, config, model):
        super().__init__()
        self.config = config
        self.final_layernorm = model.transformer.final_layernorm
    
    def forward(self, hidden_states, position_ids, attention_mask):
        hidden_states = hidden_states.permute(1, 0, 2)
        hidden_states = self.final_layernorm(hidden_states)
        return hidden_states


class ChatGLMLMHead_(nn.Module):
    def __init__(self, config, model):
        super().__init__()
        self.config = config
        self.lm_head = model.lm_head

    def forward(self, hidden_states):
        lm_logits = self.lm_head(hidden_states).permute(1, 0, 2).contiguous()
        return lm_logits



import os
from transformers import GPT2Config

def gpt_config(args):
    import json
    with open(os.path.join(os.path.dirname(__file__), 'gpt-config', '%s.json'%args.model_size)) as f:
        params = json.load(f)
    return GPT2Config(**params)

def overwrite_configs_and_args(config, args):
    overwrite_config = {'use_cache': False,
                        'use_flash_attn': args.use_flash_attn,
                        'fused_bias_fc': True,
                        'sequence_parallel': False}
    for key, val in overwrite_config.items():
        setattr(config, key, val)
    
    if args.overwrite_config:
        overwrite_config = {'hidden_size': args.hidden_size,
                            'max_position_embeddings': args.seq_length,
                            'num_hidden_layers': args.num_hidden_layers,
                            'vocab_size': args.vocab_size}
        for key, val in overwrite_config.items():
            setattr(config, key, val)
    else:
        args.hidden_size = config.hidden_size
        args.seq_length = config.max_position_embeddings
        args.max_position_embeddings = config.max_position_embeddings
        args.num_hidden_layers = config.num_hidden_layers
        args.vocab_size = config.vocab_size


import sys
import numpy as np
from ..utils import strategy2config, read_json_config, write_json_config, read_allreduce_bandwidth_config, read_p2p_bandwidth_config, estimate_bsz_start_16gpus
import argparse
import re

def match_str(s, pattern):
    match = re.match(pattern, s)
    layernum = int(match.group(1))
    bsz = int(match.group(2))
    rank = int(match.group(3))
    type = match.group(4)
    return layernum, bsz, rank, type

def process_profiled_memory(args):
    world_size = args.gpu_num
    fp16 = '_fp16' if args.use_fp16 else ('_bf16' if args.use_bf16 else '')
    memory_config_path = './configs/memory_profiling_%dgpus_dist%s_hidden%d_head%d_seqlen%d.json'%(world_size, fp16, args.hidden_size, args.num_attention_heads, args.seq_length)
    config = read_json_config(memory_config_path)
    if args.use_fp16:
        memory_config_fp16_path = './configs/memory_profiling_%dgpus_dist_fp16_hidden%d_head%d_seqlen%d.json'%(world_size, args.hidden_size, args.num_attention_heads, args.seq_length)
        config_fp16 = read_json_config(memory_config_fp16_path)
    pattern = r'layernum(\d+)_bsz(\d+)_rank(\d+)_(\w+)'
    format = "layernum%d_bsz%d_rank%d_%s"
    param_result, act_result, param = dict(), dict(), -1
    bsz = 8

    pp_deg, tp_deg = 1, 1
    while True:
        if pp_deg * tp_deg > world_size:
            break
        print(pp_deg, tp_deg)
        strategy = '%d_%d_%d'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
        if strategy not in config:
            tp_deg *= 2
            continue
        re = config[strategy]
        if args.use_fp16:
            re_fp16 = config_fp16[strategy]
        layernums = []
        for key, val in re.items():
            # print(key, val)
            layernum, bsz, rank, type = match_str(key, pattern)
            layernums.append(layernum)
        layernums = sorted(list(set(layernums)))
        param_per_layer = (re[format%(layernums[1], bsz, 0, 'ms')] - re[format%(layernums[0], bsz, 0, 'ms')])/(layernums[1]-layernums[0])*pp_deg/4
        act_per_layer_per_sample = (re[format%(layernums[1], bsz, 0, 'act')] - re[format%(layernums[0], bsz, 0, 'act')])/(layernums[1]-layernums[0])*pp_deg/(pp_deg*tp_deg)
        if args.dp_type == 'zero3':
            param_per_layer *= world_size//pp_deg//tp_deg
        if args.use_fp16:
            act_per_layer_per_sample = (re_fp16[format%(layernums[1], bsz, 0, 'act')] - re_fp16[format%(layernums[0], bsz, 0, 'act')])/(layernums[1]-layernums[0])*pp_deg/(pp_deg*tp_deg)
        param = max(param, param_per_layer*tp_deg)
        print(param_per_layer, act_per_layer_per_sample, param)
        param_result[tp_deg] = param_per_layer
        act_result[tp_deg] = act_per_layer_per_sample
        tp_deg *= 2

    print('param:', param)
    # print('param_dict:', param_result)
    print('act_dict:', act_result)

    act_dict_c = dict()
    pp_deg, tp_deg = 1, 1
    act_cpt = -1
    while True:
        if pp_deg * tp_deg > world_size:
            break
        print(pp_deg, tp_deg)
        strategy = '%d_%d_%d_c'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
        if strategy not in config:
            tp_deg *= 2
            continue
        re = config[strategy]
        if args.use_fp16:
            re = config_fp16[strategy]
        layernums = []
        for key, val in re.items():
            # print(key, val)
            layernum, bsz, rank, type = match_str(key, pattern)
            layernums.append(layernum)
        layernums = sorted(list(set(layernums)))
        act_per_layer_per_sample = (re[format%(layernums[1], bsz, 0, 'act')] - re[format%(layernums[0], bsz, 0, 'act')])/(layernums[1]-layernums[0])*pp_deg/(pp_deg*tp_deg)
        print(act_per_layer_per_sample)
        act_cpt = max(act_cpt, act_per_layer_per_sample)
        act_dict_c[tp_deg] = act_per_layer_per_sample
        tp_deg *= 2

    print('act_dict_c:', act_dict_c)
    print('act_cpt:', act_cpt)
    act_result['checkpoint'] = act_cpt

    inf=1e6
    other_memory_pp_off, other_memory_pp_on_first, other_memory_pp_on_last = \
        {'model_states': inf, 'activation': inf}, {'model_states': inf, 'activation': inf}, {'model_states': inf, 'activation': inf}
    pp_deg = 1
    while True:
        if pp_deg > world_size:
            break
        tp_deg = 1
        while True:
            if pp_deg * tp_deg > world_size:
                break
            print(pp_deg, tp_deg)
            strategy = '%d_%d_%d'%(pp_deg,tp_deg,world_size//pp_deg//tp_deg)
            if strategy not in config:
                tp_deg *= 2
                continue
            re = config[strategy]
            if args.use_fp16:
                re_fp16 = config_fp16[strategy]
            layernums = []
            for key, val in re.items():
                # print(key, val)
                layernum, bsz, rank, type = match_str(key, pattern)
                layernums.append(layernum)
            layernums = sorted(list(set(layernums)))
            other_ms_first = re[format%(layernums[0], bsz, 0, 'ms')] - layernums[0] / pp_deg * param_result[tp_deg] * 4
            if args.dp_type == 'zero3':
                other_ms_first = (re[format%(layernums[0], bsz, 0, 'ms')] - layernums[0] / pp_deg * param_result[tp_deg] * 4 / (world_size//pp_deg//tp_deg)) * world_size//pp_deg
            other_act_first = re[format%(layernums[0], bsz, 0, 'act')] - layernums[0] / pp_deg * act_result[tp_deg] * (pp_deg*tp_deg)
            other_ms_last = re[format%(layernums[0], bsz, world_size-1, 'ms')] - layernums[0] / pp_deg * param_result[tp_deg] * 4
            if args.dp_type == 'zero3':
                other_ms_last = (re[format%(layernums[0], bsz, world_size-1, 'ms')] - layernums[0] / pp_deg * param_result[tp_deg] * 4 / (world_size//pp_deg//tp_deg)) * world_size//pp_deg
            other_act_last = re[format%(layernums[0], bsz, world_size-1, 'act')] - layernums[0] / pp_deg * act_result[tp_deg] * (pp_deg*tp_deg)
            if args.use_fp16:
                other_act_first = re_fp16[format%(layernums[0], bsz, 0, 'act')] - layernums[0] / pp_deg * act_result[tp_deg] * (pp_deg*tp_deg)
                other_act_last = re_fp16[format%(layernums[0], bsz, world_size-1, 'act')] - layernums[0] / pp_deg * act_result[tp_deg] * (pp_deg*tp_deg)
            print(other_ms_first, other_act_first, other_ms_last, other_act_last)
            if pp_deg == 1:
                other_memory_pp_off['model_states'] = min(other_memory_pp_off['model_states'], other_ms_first)
                other_memory_pp_off['activation'] = min(other_memory_pp_off['activation'], other_act_first)
            else:
                other_memory_pp_on_first['model_states'] = min(other_memory_pp_on_first['model_states'], other_ms_first)
                other_memory_pp_on_first['activation'] = min(other_memory_pp_on_first['activation'], other_act_first / pp_deg)
                other_memory_pp_on_last['model_states'] = min(other_memory_pp_on_last['model_states'], other_ms_last)
                other_memory_pp_on_last['activation'] = min(other_memory_pp_on_last['activation'], other_act_last / pp_deg)
            tp_deg *= 2
        pp_deg *=2

    # other_memory_pp_on_first['activation'] = other_memory_pp_on_last['activation'] = max(other_memory_pp_on_first['activation'], other_memory_pp_on_last['activation'])
    print('other_memory_pp_off:', other_memory_pp_off)
    print('other_memory_pp_on_first:', other_memory_pp_on_first)
    print('other_memory_pp_on_last:', other_memory_pp_on_last)

    if args.use_fp16:
        config_fp16['parameter_size'] = param
        config_fp16['tp_activation_per_bsz_dict'] = act_result
        config_fp16['other_memory_pp_off'] = other_memory_pp_off
        config_fp16['other_memory_pp_on_first'] = other_memory_pp_on_first
        config_fp16['other_memory_pp_on_last'] = other_memory_pp_on_last
        write_json_config(config_fp16, memory_config_fp16_path)
    else:
        config['parameter_size'] = param
        config['tp_activation_per_bsz_dict'] = act_result
        config['other_memory_pp_off'] = other_memory_pp_off
        config['other_memory_pp_on_first'] = other_memory_pp_on_first
        config['other_memory_pp_on_last'] = other_memory_pp_on_last
        write_json_config(config, memory_config_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gpu_num", type=int, default=8, help="Number of GPUs",
    )
    parser.add_argument(
        "--hidden_size", type=int, default=768, help="Hidden size of transformer model",
    )
    parser.add_argument(
        "-a",
        "--num_attention_heads",
        type=int,
        default=12,
        help="Number of attention heads",
    )
    parser.add_argument(
        "-s", "--seq_length", type=int, default=128, help="Maximum sequence len"
    )
    parser.add_argument(
        "--use_fp16", type=int, default=0, help="Galvatron Run model in fp16 mode."
    )
    parser.add_argument(
        "--use_bf16", type=int, default=0, help="Galvatron Run model in bf16 mode."
    )
    parser.add_argument(
        "--dp_type", type=str, default='ddp', help="Use ddp or zero3 to profile memory.", choices=['ddp', 'zero3']
    )
    args = parser.parse_args()
    process_profiled_memory(args)


from .dataloader import DataLoaderForGPT
from .hybrid_parallel_model_dist import overwrite_megatron_args, get_hybrid_parallel_configs, construct_hybrid_parallel_model
from .train_hp_layerwise_dist import forward_step_func